"""Geodetic transformations. This module holds many functions to peform common
transformations used in astrodynamics.

Copyright (C) {2017}  {Shankar Kulumani}

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
import numpy as np
from kinematics import attitude
from astro import time, constants, transform
import pdb
# TODO: Make these funcitons vectorized to handle vector inputs


def lla2ecef(lat, lon, alt, r=6378.137, ee=8.1819190842622e-2):
    """Convert latitude, longitude, and altitude to the Earth centered
    Earth fixed frame (ECEF)

    Will convert between geodetic latitude, longitude, and altitude to the
    associated planet fixed reference frame.  This assumes an oblate spheroid
    model for the planetary body. It is applicable to any spherical body given
    the semi-major axis and the body eccentricity constant.

    Parameters:
    -----------
    lat : float
        Geodetic latitude (radians)
    lon : float
        Geodetic longitude (radians)
    alt : float
        height about the mean surface ellipsoid (WGS84 for Earth) (kilometers)

    Returns
    -------
    ecef : float ndarray (3,)
        Body fixed position vector in (kilometers)

    Notes
    -----
    We're assuming an oblate spheroid model for the Earth. To transform this
    vector to the inertial ECI frame all you need to do is rotate it by the
    Greenwich Mean Sideral time (GST) or use a fancy ECEF-ECI rotation which
    accounts for the Earth's precession/nutation.

    References
    ----------
    .. [1] BATE, Roger R, MUELLER, Donald D WHITE, Jerry E. Fundamentals of
    Astrodynamics. Courier Dover Publications, 1971.

    .. [2] Nice website to verify computations:
    http://www.oc.nps.edu/oc2902w/coord/llhxyz.htm

    Examples
    --------
    Some examples demonstrating the usage of this function

    >>> import numpy

    Convert the lattitude, longitude, and altitude of Washington, DC to it's
    equivalent ECEF representation.

    >>> lat, lon, alt = (38.8895 * np.pi / 180, -77.035 * np.pi / 180, 0.054)

    We need to make sure the inputs are in the correct units.

    >>> lla2ecef(lat, lon, alt)
    [1115.308, -4844.546, 3982.965]

    """
    # Normal distance from teh surface to the Z axis along the ellipsoid normal
    N = r / np.sqrt(1 - ee**2 * np.sin(lat)**2)

    x = (N + alt) * np.cos(lat) * np.cos(lon)
    y = (N + alt) * np.cos(lat) * np.sin(lon)
    z = ((1 - ee**2) * N + alt) * np.sin(lat)

    return np.array([x, y, z])



# TODO: Change documentation to ECI (inertial frame) and verify


def site2eci(lat, alt, lst, r=6378.137, ee=8.1819190842622e-2):
    """Calculate the site vector in the ECI coordinate system.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sitlat - site latitude (radians)
        lst - local sidereal time (radians)
        sitalt - site altitude (meters)

    Outputs:
        R_site - site vector in ECI frame - inertial frame

    Globals:
        RE, EEsqrd

    Constants: None

    Coupling: None

    References:
        Astro 321 COMFIX
        Vallado Alg 47
    """
    N = r / np.sqrt(1 - ee**2 * np.sin(lat)**2)
    x = (N + alt) * np.cos(lat)
    z = ((1 - ee**2) * N + alt) * np.sin(lat)
    eci = np.array([x * np.cos(lst), x * np.sin(lst), z])

    return eci

# TODO:Add documentation


def ecef2lla(ecef, r=6378.137, ee=8.1819190842622e-2):
    """Converts a ECEF vector to the equivalent Lat, longitude and Altitude
    above the reference ellipsoid

    """
    twopi = 2 * np.pi
    tol = 1e-6

    norm_vec = np.linalg.norm(ecef)

    temp = np.sqrt(ecef[0]**2 + ecef[1]**2)
    if temp < tol:
        rtasc = np.sign(ecef[2]) * np.pi * 0.5
    else:
        rtasc = np.arctan2(ecef[1], ecef[0])

    lon = rtasc
    lon = attitude.normalize(lon, 0, 2 * np.pi)

    decl = np.arcsin(ecef[2] / norm_vec)
    latgd = decl

    # iterate to find geodetic latitude
    i = 1
    olddelta = latgd + 10.0
    while np.absolute(olddelta - latgd) >= tol and i < 10:
        oldelta = latgd
        sintemp = np.sin(latgd)
        c = r / np.sqrt(1.0 - ee**2 * sintemp**2)
        latgd = np.arctan2(ecef[2] + c * ee**2 * sintemp, temp)
        i = i + 1

    # calculate the height
    if (np.pi / 2 - np.absolute(latgd)) > np.pi / 180:
        hellp = temp / np.cos(latgd) - c
    else:
        s = c * (1 - ee**2)
        hellp = ecef[2] / np.sin(latgd) - s

    latgc = gd2gc(latgd, ee**2)
    return latgc, latgd, lon, hellp


def gd2gc(latgd, eesqrd=0.081819221456**2):
    """Only valid for locations on the Earth's surface

    Vallado Example 3-1
    """
    latgc = np.arctan((1 - eesqrd) * np.tan(latgd))
    return latgc


def gc2gd(latgc, eesqrd=0.081819221456**2):
    """Only valid for locations on the Earth's surface

    Vallado Example 3-1
    """
    latgd = np.arctan(np.tan(latgc) / (1 - eesqrd))
    return latgd

# TODO: Incorporate all the other observation angles (RADEC, etc. Vallado Ch 4)


def rv2rhoazel(r_sat_eci, v_sat_eci, lat, lon, alt, jd):
    """
    This function calculates the topcentric range,azimuth and elevation from
    the site vector and satellite position vector.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sat_eci - satellite ECI position vector (km)
        site_eci - site ECI position vector (km)
        site_lat - site geodetic latitude (radians)
        site_lst - site local sidereal time (radians)

    Outputs:
        rho - range (km)
        az - asimuth (radians)
        el - elevation (radians)

    Globals: None

    Constants: None

    Coupling:

    References:
        Astro 321 Predict LSN 22
        Vallado Algorithm 27
    """
    small = constants.small
    halfpi = constants.halfpi

    # get site in ECEF
    r_site_ecef = lla2ecef(lat, lon, alt)

    # convert sat and site to ecef
    dcm_eci2ecef = transform.dcm_eci2ecef(jd)
    omega_earth = np.array([0, 0, constants.earth.omega])

    r_sat_ecef = dcm_eci2ecef.dot(r_sat_eci)
    v_sat_ecef = dcm_eci2ecef.dot(
        v_sat_eci) - np.cross(omega_earth, r_sat_ecef)

    # find relative vector in ecef frame
    rho_ecef = r_sat_ecef - r_site_ecef
    drho_ecef = v_sat_ecef - np.zeros_like(v_sat_ecef)  # site isn't moving
    rho = np.linalg.norm(rho_ecef)

    # convert to SEZ coordinate frame
    dcm_ecef2sez = attitude.rot2(
        np.pi / 2 - lat, 'r').dot(attitude.rot3(lon, 'r'))
    rho_sez = dcm_ecef2sez.dot(rho_ecef)
    drho_sez = dcm_ecef2sez.dot(drho_ecef)

    # find azimuth and eleveation
    temp = np.sqrt(rho_sez[0]**2 + rho_sez[1]**2)

    if temp < small:  # directly over the north pole
        el = np.sign(rho_sez[2]) * halfpi  # \pm 90 deg
    else:
        mag_rho_sez = np.linalg.norm(rho_sez)
        el = np.arcsin(rho_sez[2] / mag_rho_sez)

    if temp < small:
        az = np.arctan2(drho_sez[1], - drho_sez[0])
    else:
        az = np.arctan2(rho_sez[1] / temp, -rho_sez[0] / temp)

    # range, azimuth and elevation rates
    drho = np.dot(rho_sez, drho_sez) / rho

    if np.absolute(temp**2) > small:
        daz = (drho_sez[0] * rho_sez[1] - drho_sez[1] * rho_sez[0]) / temp**2
    else:
        daz = 0

    if np.absolute(temp) > small:
        dele = (drho_sez[2] - drho * np.sin(el)) / temp
    else:
        dele = 0

    az = attitude.normalize(az, 0, constants.twopi)
    return rho, az, el, drho, daz, dele


def rhoazel(sat_eci, site_eci, site_lat, site_lst):
    """
    This function calculates the topcentric range,azimuth and elevation from
    the site vector and satellite position vector.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        sat_eci - satellite ECI position vector (km)
        site_eci - site ECI position vector (km)
        site_lat - site geodetic latitude (radians)
        site_lst - site local sidereal time (radians)

    Outputs:
        rho - range (km)
        az - asimuth (radians)
        el - elevation (radians)

    Globals: None

    Constants: None

    Coupling: 

    References:
        Astro 321 Predict LSN 22 
    """

    site2sat_eci = sat_eci - site_eci

    site2sat_sez = attitude.rot3(-site_lst).dot(site2sat_eci)
    site2sat_sez = attitude.rot2(-(np.pi / 2 - site_lat)).dot(site2sat_sez)

    rho = np.linalg.norm(site2sat_sez)
    el = np.arcsin(site2sat_sez[2] / rho)
    az = attitude.normalize(np.arctan2(site2sat_sez[1], -site2sat_sez[0]), 0,
                            2 * np.pi)[0]

    return rho, az, el


def rhoazel2sez(rho, az, el, drho, daz, dele):
    """This program calculates the range vector in the SEZ system.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        rho - range (km)
        az - azimuth (radians)
        el - elevation (radians)
        drho - range rate (km/sec)
        daz - azimuth rate (radians/sec)
        del - elevation rate (radians/sec)

    Outputs:
        Rho_sez - range vector SEZ (km)
        Drho_sez - range velocity vecotr SEZ (km/sec)

    Globals: None

    Constants: None

    Coupling: None

    References:
        Astro 321 COMFIX
    """

    # Calculate range vector
    rho_s = -rho * np.cos(az) * np.cos(el)
    rho_e = rho * np.sin(az) * np.cos(el)
    rho_z = rho * np.sin(el)

    rho_sez = np.array([rho_s, rho_e, rho_z])

    # Calculate range rate vector
    transform = np.array([[-np.cos(el) * np.cos(az), rho * np.sin(el) * np.cos(az), rho * np.cos(el) * np.sin(az)],
                          [np.cos(el) * np.sin(az), -rho * np.sin(el)
                           * np.sin(az), rho * np.cos(el) * np.cos(az)],
                          [np.sin(el), rho * np.cos(el), 0]])
    drho_sez = transform.dot(np.array([drho, dele, daz]))
    # drho_s=-drho*np.cos(az)*np.cos(el)+rho*daz*np.sin(az)*np.cos(el)+rho*dele*np.cos(az)*np.sin(el)
    # drho_e=drho*np.sin(az)*np.cos(el)+rho*daz*np.cos(az)*np.cos(el)-rho*dele*np.sin(az)*np.sin(el)
    # drho_z=drho*np.sin(el)+rho*dele*np.cos(el)

    # drho_sez=np.array([drho_s,drho_e,drho_z])

    return rho_sez,  drho_sez


def eci2lla(pos_eci, jd):
    """Find LLA for a ECI vector about the Earth
    """
    pass



import numpy as np
from astro import time

deg2rad = np.pi / 180
rad2deg = 180 / np.pi

def test_julian_day_j2000():
    jd, mjd = time.date2jd(2000, 1, 1, 12, 0, 0)
    np.testing.assert_almost_equal(jd,2451545)

def test_julian_day_current():
    jd, mjd = time.date2jd(2017, 6, 10, 0, 0, 0)
    np.testing.assert_equal(jd, 2457914.5)

def test_julian_day_upper_limit():
    jd, mjd = time.date2jd(2100, 2, 28, 0, 0, 0)
    np.testing.assert_equal(jd, 2488127.5)

def test_julian_day_lower_limit():
    jd, mjd = time.date2jd(1900, 3, 1, 0, 0, 0)
    np.testing.assert_equal(jd, 2415079.5)

def test_finddays_newyears():
    ddays = time.finddays(2017, 1, 1, 0, 0, 0)
    np.testing.assert_equal(ddays, 1)

def test_finddays_leapday():
    dday = time.finddays(2000, 2, 29, 0, 0, 0)
    np.testing.assert_equal(dday, 31+29)

def test_finddays_vallado_311():
    expected_days = 129.0
    actual_days = time.finddays(1992, 5, 8, 0, 0, 0)
    np.testing.assert_allclose(actual_days, expected_days)

def test_dayofyr2mdhms_vallado_311():
    actual_yr, actual_days = 1992, 129.0
    actual_mo, actual_day, actual_hr, actual_min, actual_sec = 5, 8, 0, 0, 0

    actual_mdhms = time.dayofyr2mdhms(1992, 129)
    np.testing.assert_allclose(actual_mdhms, (actual_mo, actual_day, actual_hr, actual_min, actual_sec))

def test_vallado_312_converting_fractional_days():
    expected_days = 77.5097222
    actual_days = time.finddays(2001, 3, 18, 12, 14, 0)
    np.testing.assert_allclose(actual_days, expected_days)

def test_vallado_312_dayofyear():
    expected_mdhms = (3, 18, 12, 13, 59.99808)
    actual_mdhms = time.dayofyr2mdhms(2001, 77.5097222)
    np.testing.assert_allclose(actual_mdhms, expected_mdhms)

def test_finddays_special_year():
    expected_days = 365
    actual_days = time.finddays(1800, 12, 31, 0, 0, 0)
    np.testing.assert_allclose(actual_days, expected_days)

# TODO Finish adding dates from table 3-3
class TestDayOfLeapYear():
    # Table 3-3 from Vallado
    def test_jan_1(self):
        yrmdhms = (2000, 1, 1, 0, 0, 0)
        days = time.finddays(yrmdhms[0], yrmdhms[1], yrmdhms[2], yrmdhms[3], yrmdhms[4], yrmdhms[5])
        mdhms = time.dayofyr2mdhms(yrmdhms[0], 1)

        np.testing.assert_allclose(days, 1)
        np.testing.assert_allclose(mdhms, yrmdhms[1:])
    def test_jan_31(self):
        days = time.finddays(2000, 1, 31, 0, 0, 0)
        np.testing.assert_allclose(days, 31)

def test_date2jd_vallado_p409():
    # example from 7-1 Vallado pg. 409
    jd, mjd = time.date2jd(1995, 5, 20, 3, 17, 2)
    np.testing.assert_allclose(jd, 2449857.636829, rtol=1e-5)

def test_jd2date_vallado_p409():
    expected_date = (1995, 5, 20, 3, 17, 2.0255)
    actual_date = time.jd2date(2449857.636829)
    np.testing.assert_allclose(actual_date, expected_date, rtol=1e-3)

def test_jd2date_1900():
    expected_date = (1899, 12, 31, 0, 0, 0)
    actual_date = time.jd2date(2415019.5)
    np.testing.assert_allclose(actual_date, expected_date, rtol=1e-4)

def test_dayofyr2mdhms_leap():
    yr = 2000
    days = 31+29
    expected_date = (2, 29, 0, 0, 0)
    actual_output = time.dayofyr2mdhms(yr, days)
    np.testing.assert_allclose(actual_output, expected_date)

def test_dayofyr2mdhms_january():
    yr = 2017
    days = 1.5

    expected_output = (1, 1, 12, 0, 0)
    actual_output = time.dayofyr2mdhms(yr, days)
    np.testing.assert_array_almost_equal(actual_output, expected_output)

def test_dayofyr2mdhms_start():
    yr = 2006
    day = 170
    expected_output = (6, 19, 0, 0, 0)
    actual_output = time.dayofyr2mdhms(yr, day)
    np.testing.assert_allclose(actual_output, expected_output)

def test_dayofyr2mdhms_end():
    yr = 2006
    day = 180
    expected_output = (6, 29, 0, 0, 0)
    actual_output = time.dayofyr2mdhms(yr, day)
    np.testing.assert_allclose(actual_output, expected_output)

def test_date2jd_start():
    expected_output = 2453905.5
    actual_output = time.date2jd(2006, 6, 19, 0, 0, 0)
    np.testing.assert_allclose(actual_output[0], expected_output)

def test_date2jd_end():
    expected_output = 2453915.5
    actual_output = time.date2jd(2006, 6, 29, 0, 0, 0)
    np.testing.assert_allclose(actual_output[0], expected_output)

def test_date2jd_vallado():
    expected_jd = 2450383.09722222
    actual_output, _ = time.date2jd(1996, 10, 26, 14, 20, 0)
    np.testing.assert_allclose(actual_output, expected_jd)

def test_gst_vallado():
    expected_gst = 152.578787886
    actual_gst, _ = time.gstlst(2448855.009722, -104*deg2rad)
    np.testing.assert_allclose(actual_gst * rad2deg, expected_gst, rtol=1e-3)

def test_lst_vallado():
    expected_lst = 48.578787886
    _, actual_lst = time.gstlst(2448855.009722, -104 * deg2rad)
    np.testing.assert_allclose(actual_lst * rad2deg, expected_lst, rtol=1e-3)

class TestTimeGSTValladoEx3_3():
    dateexp = (1995, 2, 24, 12, 0, 0)
    jdexp = 2449773.0
    gstexp = np.deg2rad(333.893486)
    jd, mjd = time.date2jd(dateexp[0], dateexp[1], dateexp[2], dateexp[3], dateexp[4], dateexp[5])
    date = time.jd2date(jdexp)
    gst, _ = time.gstlst(jd, 0)

    def test_jd(self):
        np.testing.assert_allclose(self.jd, self.jdexp)

    def test_date(self):
        np.testing.assert_allclose(self.date, self.dateexp)

    def test_gst(self):
        np.testing.assert_allclose(self.gst, self.gstexp)




import numpy as np
# namedtuple to hold constants for each body
from astro import kepler, constants
import pdb

#TODO:Add test for tof_nu

def test_coe2rv_equatorial_circular():
    """Test COE to RV for equatorial circular orbit around Earth"""

    p = 6378.137 # km
    ecc = 0.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu = 0.0
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6378.137,0,0])
    V_ijk_true = np.sqrt(mu/p) * np.array([0,1,0])
    R_pqw_true = R_ijk_true
    V_pqw_true = V_ijk_true

    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p,ecc,inc,raan,arg_p,nu, mu)

    np.testing.assert_array_almost_equal(R_ijk_true,R_ijk)
    np.testing.assert_array_almost_equal(V_ijk_true,V_ijk)
    np.testing.assert_array_almost_equal(V_pqw_true,V_pqw)
    np.testing.assert_array_almost_equal(V_pqw_true,V_pqw)

class Testrv2coeEquatorialCircular():

    p_true = 6378.137 # km
    ecc_true = 0.0
    inc_true = 0.0
    raan_true = 0.0
    arg_p_true = 0.0
    nu_true = 0.0
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6378.137,0,0])
    V_ijk_true = np.sqrt(mu/p_true) * np.array([0,1,0])
    
    p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper = kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2))

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true)

    def test_arglat(self):
        np.testing.assert_allclose(self.arglat, self.nu_true + self.arg_p_true)

    def test_truelon(self):
        np.testing.assert_allclose(self.truelon, self.nu_true + self.raan_true + self.arg_p_true)

    def test_lonper(self):
        np.testing.assert_allclose(self.lonper, self.arg_p_true + self.raan_true)

def test_coe2rv_polar_circular():
    """Test COE to RV for polar circular orbit around Earth"""

    p = 6378.137 # km
    ecc = 0.0
    inc = np.pi/2
    raan = 0.0
    arg_p = 0.0
    nu = 0.0
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6378.137,0,0])
    V_ijk_true = np.sqrt(mu/p) * np.array([0.0,0.0,1])

    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p,ecc,inc,raan,arg_p,nu, mu)

    np.testing.assert_array_almost_equal(R_ijk_true,R_ijk)
    np.testing.assert_array_almost_equal(V_ijk_true,V_ijk)

class Testrv2coePolarCircular():

    p_true = 6378.137 # km
    ecc_true = 0.0
    inc_true = np.pi / 2
    raan_true = 0.0
    arg_p_true = 0.0
    nu_true = 0.0
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6378.137,0,0])
    V_ijk_true = np.sqrt(mu/p_true) * np.array([0,0,1])
    
    p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper = kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2))

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true)

    def test_arglat(self):
        np.testing.assert_allclose(self.arglat, self.nu_true + self.arg_p_true)

    def test_truelon(self):
        np.testing.assert_allclose(self.truelon, self.nu_true + self.raan_true + self.arg_p_true)

    def test_lonper(self):
        np.testing.assert_allclose(self.lonper, self.arg_p_true + self.raan_true)

def test_coe2rv_equatorial_circular_quarter():
    """Test COE to RV for equatorial circular orbit around Earth"""

    p = 6378.137 # km
    ecc = 0.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu = np.pi/2
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([0,6378.137,0])
    V_ijk_true = np.sqrt(mu/p) * np.array([-1.0,0.0,0])

    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p,ecc,inc,raan,arg_p,nu, mu)

    np.testing.assert_array_almost_equal(R_ijk_true,R_ijk)
    np.testing.assert_array_almost_equal(V_ijk_true,V_ijk)    

class Testrv2coeEquatorialCircularQuarer():

    p_true = 6378.137 # km
    ecc_true = 0.0
    inc_true = 0.0
    raan_true = 0.0
    arg_p_true = 0.0
    nu_true = np.pi /2
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([0.0, 6378.137,0])
    V_ijk_true = np.sqrt(mu/p_true) * np.array([-1,0,0])
    
    p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper = kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2))

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true)

    def test_arglat(self):
        # argument of latitude isn't used for this case so it goes to zero
        # np.testing.assert_allclose(self.arglat, self.nu_true + self.arg_p_true)
        pass

    def test_truelon(self):
        np.testing.assert_allclose(self.truelon, self.nu_true + self.raan_true + self.arg_p_true)

    def test_lonper(self):
        np.testing.assert_allclose(self.lonper, self.arg_p_true + self.raan_true)

class Testrv2coeVallado():

    p_true = 1.73527 * constants.earth.radius # km
    ecc_true = 0.83285
    inc_true = np.deg2rad(87.87)
    raan_true = np.deg2rad(227.89)
    arg_p_true = np.deg2rad(53.38)
    nu_true = np.deg2rad(92.335)
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6524.834, 6862.875, 6448.296])
    V_ijk_true = np.array([4.901320, 5.533756, -1.976341]) 
    
    p, a, ecc, inc, raan, arg_p, nu, m, _, _, _= kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    rtol = 1e-3

    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true, rtol=self.rtol)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2), rtol=self.rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true, rtol=self.rtol)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true, rtol=self.rtol)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true, rtol=self.rtol)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true, rtol=self.rtol)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true, rtol=self.rtol)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true, rtol=self.rtol)


class Testrv2coeCurtis():
    """Using example 4.7 from Curtis
    """
    p_true = 80000**2/398600
    ecc_true = 1.4
    inc_true = np.deg2rad(30)
    raan_true = np.deg2rad(40)
    arg_p_true = np.deg2rad(60)
    nu_true = np.deg2rad(30)
    mu = 398600 # km^3 /sec^2

    R_ijk_true = np.array([-4040, 4815, 3629])
    V_ijk_true = np.array([-10.39, -4.772, 1.744]) 
    
    p, a, ecc, inc, raan, arg_p, nu, m, _, _, _= kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    rtol = 1e-3

    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true, rtol=self.rtol)
    
    def test_a(self):
        np.testing.assert_allclose(np.absolute(self.a), self.p_true / (self.ecc_true**2 - 1), rtol=1e2)

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true, rtol=1e-1)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true, rtol=self.rtol)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true, rtol=self.rtol)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true, rtol=self.rtol)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true, rtol=self.rtol)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true, rtol=self.rtol)

def test_coe2rv_curtis():
    """Test COE to RV for Curtis example 4.3"""

    p_true = 80000**2/398600
    ecc_true = 1.4
    inc_true = np.deg2rad(30)
    raan_true = np.deg2rad(40)
    arg_p_true = np.deg2rad(60)
    nu_true = np.deg2rad(30)
    mu = 398600 # km^3 /sec^2

    R_ijk_true = np.array([-4040, 4815, 3629])
    V_ijk_true = np.array([-10.39, -4.772, 1.744]) 
    
    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p_true,ecc_true,inc_true,raan_true,arg_p_true,nu_true, mu)

    np.testing.assert_allclose(R_ijk,R_ijk_true, rtol=1e0)
    np.testing.assert_allclose(V_ijk,V_ijk_true, rtol=1e0)

def test_coe2rv_vallado():
    """Test COE to RV for Vallado example 2-6"""

    p_true = 1.735272 * constants.earth.radius # km
    ecc_true = 0.832853
    inc_true = np.deg2rad(87.87)
    raan_true = np.deg2rad(227.89)
    arg_p_true = np.deg2rad(53.38)
    nu_true = np.deg2rad(92.336)
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([6525.5454,6861.7313, 6449.0585])
    V_ijk_true = np.array([4.902227, 5.533085, -1.975757]) 
    
    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p_true,ecc_true,inc_true,raan_true,arg_p_true,nu_true, mu)

    np.testing.assert_allclose(R_ijk,R_ijk_true, rtol=1e-4)
    np.testing.assert_allclose(V_ijk,V_ijk_true, rtol=1e-4)

def test_coe2rv_equatorial_circular_half():
    """Test COE to RV for equatorial circular orbit around Earth"""

    p = 6378.137 # km
    ecc = 0.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu = np.pi
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([-p,0,0])
    V_ijk_true = np.sqrt(mu/p) * np.array([0.0,-1.0,0])

    R_ijk, V_ijk, R_pqw, V_pqw = kepler.coe2rv(p,ecc,inc,raan,arg_p,nu, mu)

    np.testing.assert_array_almost_equal(R_ijk_true,R_ijk)
    np.testing.assert_array_almost_equal(V_ijk_true,V_ijk)

class Testrv2coeRV1():
    a_true = 8697.5027
    ecc_true = 0.2802
    p_true = a_true * ( 1 - ecc_true**2)
    inc_true = np.deg2rad(33.9987)
    raan_true = np.deg2rad(250.0287)
    arg_p_true = np.deg2rad(255.5372)
    nu_true = np.deg2rad(214.8548)
    mu = 398600.5

    r = np.array([8840, 646, 5455])
    v = np.array([-0.6950, 5.250, -1.65])

    p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper = kepler.rv2coe(r, v, mu)

    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true, rtol=1e-1)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2))

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true, rtol=1e-4)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true, rtol=1e-4)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true, rtol=1e-4)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true, rtol=1e-4)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true)

    def test_arglat(self):
        # argument of latitude isn't used for this case so it goes to zero
        # np.testing.assert_allclose(self.arglat, self.nu_true + self.arg_p_true)
        pass

    def test_truelon(self):
        # not used for inclined orbits - only special ones
        # np.testing.assert_allclose(self.truelon, self.nu_true + self.raan_true + self.arg_p_true)
        pass

    def test_lonper(self):
        # not really used for inclined orbits
        # np.testing.assert_allclose(self.lonper, self.arg_p_true + self.raan_true)
        pass

class Testrv2coeEquatorialCircularHalf():

    p_true = 6378.137 # km
    ecc_true = 0.0
    inc_true = 0.0
    raan_true = 0.0
    arg_p_true = 0.0
    nu_true = np.pi
    mu = 398600.5 # km^3 /sec^2

    R_ijk_true = np.array([-p_true, 0,0])
    V_ijk_true = np.sqrt(mu/p_true) * np.array([0,-1,0])
    
    p, a, ecc, inc, raan, arg_p, nu, m, arglat, truelon, lonper = kepler.rv2coe(R_ijk_true, V_ijk_true, mu)
    
    def test_p(self):
        np.testing.assert_allclose(self.p, self.p_true)
    
    def test_a(self):
        np.testing.assert_allclose(self.a, self.p_true / (1 - self.ecc_true**2))

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true)
    
    def test_inc(self):
        np.testing.assert_allclose(self.inc, self.inc_true)

    def test_raan(self):
        np.testing.assert_allclose(self.raan, self.raan_true)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p, self.arg_p_true)
    
    def test_nu(self):
        np.testing.assert_allclose(self.nu, self.nu_true)

    def test_m(self):
        E_true, M_true = kepler.nu2anom(self.nu, self.ecc)
        np.testing.assert_allclose(self.m, M_true)

    def test_arglat(self):
        # argument of latitude isn't used for this case so it goes to zero
        # np.testing.assert_allclose(self.arglat, self.nu_true + self.arg_p_true)
        pass

    def test_truelon(self):
        np.testing.assert_allclose(self.truelon, self.nu_true + self.raan_true + self.arg_p_true)

    def test_lonper(self):
        np.testing.assert_allclose(self.lonper, self.arg_p_true + self.raan_true)
def test_kepler_eq_E():
    """
        A series of cases run in Matlab and copied here
    """
    M = np.deg2rad(110)
    ecc = 0.9
    E_matlab = 2.475786297687611 
    nu_matlab = 2.983273149717047

    E_python, nu_python, count_python = kepler.kepler_eq_E(M,ecc)

    np.testing.assert_allclose((E_python, nu_python),(E_matlab,nu_matlab))

def test_kepler_eq_E_zero():
    """
        Make sure that at zero nu=E=M
    """
    M = 0.0
    ecc = 1.2
    E_true = 0.0
    nu_true = 0.0
    
    E_python, nu_python, count_python = kepler.kepler_eq_E(M,ecc)

    np.testing.assert_array_almost_equal((E_python,nu_python),(E_true,nu_true))


def test_kepler_eq_E_pi():
    """
        Make sure that at pi nu=E=M
    """
    M = np.pi
    ecc = 0.9
    E_true = np.pi
    nu_true = np.pi

    E_python, nu_python, count_python = kepler.kepler_eq_E(M,ecc)

    np.testing.assert_array_almost_equal((E_python,nu_python),(E_true,nu_true))

def test_nu2anom():
    """
        A matlab test case which is copied here
    """
    M_matlab = np.deg2rad(110)
    ecc = 0.9
    E_matlab = 2.475786297687611 
    nu_matlab = 2.983273149717047

    E_python, M_python = kepler.nu2anom(nu_matlab,ecc)

    np.testing.assert_allclose(E_python, E_matlab)
    np.testing.assert_allclose(M_python, M_matlab)

def test_nu2anom_zero():
    """
        Make sure that at zero nu=E=M
    """
    M_true = 0.0
    ecc = 1.2
    E_true = 0.0
    nu_true = 0.0

    E_python, M_python = kepler.nu2anom(nu_true,ecc)

    np.testing.assert_allclose((E_python, M_python),(E_true,M_true))


def test_nu2anom_pi():
    """
        Make sure that at pi nu=E=M
    """
    M_true = np.pi
    ecc = 0.9
    E_true = np.pi
    nu_true = np.pi
    
    E_python, M_python = kepler.nu2anom(nu_true,ecc)

    np.testing.assert_allclose(E_python, E_true)
    np.testing.assert_allclose(M_python, M_true)


def test_tof_delta_t():
    """Test propogation using Kepler's Eq"""

    # define circular orbit around the Earth
    p = 8000
    ecc = 0
    mu = constants.earth.mu
    nu_0 = np.pi/2
    delta_t = 2 * np.pi * np.sqrt(p**3/ mu)
    
    # make sure you get back to the same spot
    E_f, M_f, nu_f = kepler.tof_delta_t( p, ecc, mu, nu_0, delta_t)
    np.testing.assert_allclose(nu_f, nu_0)

    

def test_fpa_solve_circular():
    fpa_actual = kepler.fpa_solve(0, 0)
    fpa_expected = 0 
    np.testing.assert_allclose(fpa_actual, fpa_expected)

def test_fpa_solve_elliptical():
    fpa_actual = kepler.fpa_solve(0, 0.5)
    fpa_expected = 0
    np.testing.assert_allclose(fpa_actual, fpa_expected)

def test_fpa_solve_parabolic():
    fpa_actual = kepler.fpa_solve(0, 1)
    fpa_expected = 0
    np.testing.assert_allclose(fpa_actual, fpa_expected)

def test_fpa_solve_hyperbolic():
    fpa_actual = kepler.fpa_solve(0, 2)
    fpa_expected = 0
    np.testing.assert_allclose(fpa_actual, fpa_expected)

class TestConicOrbitEquatorial():

    p = 7000 # km
    ecc = 0.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu_i = 0.0
    nu_f = 0.0 
    mu = constants.earth.mu
    (state_eci, state_pqw, state_lvlh, state_sat_eci, state_sat_pqw,
     state_sat_lvlh) = kepler.conic_orbit(p, ecc, inc, raan, arg_p, nu_i, nu_f,
                                          mu)

    def test_x_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[0], self.p)

    def test_y_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[1], 0)

    def test_z_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[2], 0)

    def test_circular_orbit(self):
        np.testing.assert_allclose(np.linalg.norm(self.state_eci[:, 0:3],
                                                  axis=1), self.p)
    
    def test_lvlh_radial_velocity(self):
        np.testing.assert_allclose(self.state_lvlh[:, 3], 0)

    def test_lvlh_tangential_velocity(self):
        np.testing.assert_allclose(self.state_lvlh[:, 4], np.sqrt(self.mu/self.p))

class TestConicOrbitHyperbolic():

    p = 10000 # km
    ecc = 2.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu_i = 0.0
    nu_f = 0.0 

    (state_eci, state_pqw, state_lvlh, state_sat_eci, state_sat_pqw,
     state_sat_lvlh) = kepler.conic_orbit(p, ecc, inc, raan, arg_p, nu_i, nu_f)
    
    def test_x_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[0], self.p/(1-self.ecc**2)*(1-self.ecc))

    def test_y_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[1], 0)

    def test_z_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[2], 0)

class TestConicOrbitParabolic():

    p = 10000 # km
    ecc = 1.0
    inc = 0.0
    raan = 0.0
    arg_p = 0.0
    nu_i = 0.0
    nu_f = 0.0 

    (state_eci, state_pqw, state_lvlh, state_sat_eci, state_sat_pqw,
     state_sat_lvlh) = kepler.conic_orbit(p, ecc, inc, raan, arg_p, nu_i, nu_f)
    
    def test_x_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[0], self.p/2)

    def test_y_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[1], 0)

    def test_z_axis(self):
        np.testing.assert_allclose(self.state_sat_eci[2], 0)

class TestEllipticalOribtProperties():
    # test case from RV2COE Astro 321, MAE3145
    r = np.array([8840.0, 646, 5455])
    v = np.array([-0.695, 5.25, -1.65])

    r_mag_true = 10407.6866
    v_mag_true = 5.5469

    r_per_true = 6260.5311
    r_apo_true = 11134.4744
    energy_true = -22.9147
    period_true = 2.2423
    sma_true = 8697.5027
    ecc_true = 0.2802
    inc_true = 33.9987
    raan_true = 250.0287
    arg_p_true = 255.5372
    nu_true = 214.8548
    
    mu = constants.earth.mu

    p, a, ecc, inc, raan, arg_p, nu, _, _, _, _ = kepler.rv2coe(r, v, mu) 
    ( a, h, period, sme, fpa, r_per, r_apo, r_ijk, v_ijk,
     r_pqw, v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc,
     E, M, n ) = kepler.elp_orbit_el(p,ecc,inc,raan,arg_p,nu,mu)
    def test_r_mag(self):
        np.testing.assert_allclose(np.linalg.norm(self.r_ijk), self.r_mag_true, rtol=1e-4)

    def test_v_mag(self):
        np.testing.assert_allclose(np.linalg.norm(self.v_ijk), self.v_mag_true, rtol=1e-4)

    def test_r_per(self):
        np.testing.assert_allclose(self.r_per, self.r_per_true, rtol=1e-4)

    def test_r_apo(self):
        np.testing.assert_allclose(self.r_apo, self.r_apo_true, rtol=1e-4)

    def test_energy(self):
        np.testing.assert_allclose(self.sme, self.energy_true, rtol=1e-4)

    def test_period(self):
        np.testing.assert_allclose(self.period*constants.sec2hr, self.period_true, rtol=1e-4)

    def test_sma(self):
        np.testing.assert_allclose(self.a, self.sma_true, rtol=1e-4)

    def test_ecc(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true, rtol=1e-4)

    def test_inc(self):
        np.testing.assert_allclose(self.inc*constants.rad2deg, self.inc_true, rtol=1e-4)

    def test_raan(self):
        np.testing.assert_allclose(self.raan*constants.rad2deg, self.raan_true, rtol=1e-4)

    def test_arg_p(self):
        np.testing.assert_allclose(self.arg_p*constants.rad2deg, self.arg_p_true, rtol=1e-4)

class TestHyperbolicOrbitProperties():
    """From MAE3145 HW4 Problem 2
    """
    mu = constants.earth.mu
    rp = 1000 + constants.earth.radius
    ecc = 1.05

    a_true = 147562.7399
    p_true = 15125.19
    energy_true = 1.350613644
    vinf_true = 1.64354108
    nu_inf_true = np.deg2rad(162.2472)
    flyby_true = np.deg2rad(144.4944196)

    a, p = kepler.hyp_per2sma(rp, ecc) 
    (a, v_inf, b, sme, flyby, nu_inf, h, fpa, r_per, r_ijk, v_ijk, r_pqw,
     v_pqw, r_lvlh, v_lvlh, r, v, v_circ, v_esc, H, M_H, n) = kepler.hyp_orbit_el(p, ecc, 0, 0, 0, np.pi/2, mu)
   
    def test_semi_major_axis(self):
        np.testing.assert_allclose(self.a, self.a_true)
    
    def test_specific_mechanical_energy(self):
        np.testing.assert_allclose(self.sme, self.energy_true)
    
    def test_velocity_infinity(self):
        np.testing.assert_allclose(self.v_inf, self.vinf_true)
    
    def test_true_anomaly_infinity(self):
        np.testing.assert_allclose(self.nu_inf, self.nu_inf_true)
    
    def test_flyby_angle(self):
        np.testing.assert_allclose(self.flyby, self.flyby_true)

class TestHNEVector_elliptical_equatorial():
    r, v, _, _ = kepler.coe2rv(10000, 0.2, 0, 0, 0, 0, constants.earth.mu)
    h, n, e = kepler.hne_vec(r, v, constants.earth.mu)

    def test_h_vec(self):
        np.testing.assert_allclose(self.h, np.array([0, 0, 1]))

    def test_n_vec(self):
        np.testing.assert_allclose(self.n, np.zeros(3))

    def test_e_vec(self):
        np.testing.assert_allclose(self.e, np.array([1, 0, 0]))
        
def test_perapo2aecc_circular():
    r_per = 8000
    r_apo = r_per
    a_actual = r_per
    p_actual = r_per
    ecc_actual = 0
    a, p, ecc = kepler.perapo2aecc(r_per, r_apo)
    np.testing.assert_allclose((a, p, ecc), (a_actual, p_actual, ecc_actual))

def test_perapo2aecc_semi_major_axis_mean():
    r_per = 10000
    r_apo = 20000
    a_actual = (r_per + r_apo) / 2
    a, _, _ = kepler.perapo2aecc(r_per, r_apo)
    np.testing.assert_allclose(a, a_actual)

class TestSemiLatusRectum():

    def test_circular(self):
        a = 8000
        ecc = 0
        p = kepler.semilatus_rectum(a, ecc)
        np.testing.assert_allclose(p, a)
    
    def test_elliptical(self):
        a = 8000
        ecc = 0.5
        p = kepler.semilatus_rectum(a, ecc)
        np.testing.assert_allclose(p, a * (1 - ecc**2))

    def test_parabolic(self):
        """Shouldn't work since parabolas have infinite a
        """
        a = np.infty
        ecc = 1
        p = kepler.semilatus_rectum(a, ecc)
        np.testing.assert_allclose(p, 0)
    
    def test_hyperbolic(self):
        a = -8000
        ecc = 1.5
        p = kepler.semilatus_rectum(a, ecc)
        np.testing.assert_allclose(p, a * (1 - ecc**2))

def test_hyperbolic_semi_major_axis():
    a_actual = -147562.739
    p_actual = 15125.180

    a, p = kepler.hyp_per2sma(1000 + 6378.137, 1.05)
    np.testing.assert_allclose(a, a_actual)
    np.testing.assert_allclose(p, p_actual)

def test_true_anomaly_solve():
    a_1 =6*6378.137
    ecc_1 = .5
    nu_0 = 0
    p_1 = a_1*(1-ecc_1**2)
    nu = kepler.nu_solve(p_1, ecc_1, 7.6 * 6378.137)
    np.testing.assert_allclose(nu[0], np.deg2rad(144.6654977))



"""Test the transform module
"""

from astro import transform, time, geodetic
from kinematics.attitude import rot2, rot3
from kinematics import attitude
import numpy as np

# TODO: Add more test functions - from book
class TestPQWtoECI():
    raan = np.deg2rad(227.89)
    inc = np.deg2rad(87.87)
    argp = np.deg2rad(53.38)

    r = np.array([6524.834, 6862.875, 6448.296])
    v = np.array([4.901320, 5.533756, -1.976341])

    def test_both_functions_equivalent(self):
        dcm_pqw2eci_vector = transform.dcm_pqw2eci_vector(self.r, self.v)
        dcm_pqw2eci_coe = transform.dcm_pqw2eci_coe(self.raan, self.inc, self.argp)
        np.testing.assert_allclose(dcm_pqw2eci_coe, dcm_pqw2eci_vector, rtol=1e-3)

class TestECEF2ENU():

    def test_equator_prime_meridian(self):
        latgd = 0
        lon = 0
        alt = 0
        dcm_ecef2enu_expected = np.array([[0, 0, 1],
                                          [1, 0, 0],
                                          [0, 1, 0]]).T
        dcm_ecef2enu = transform.dcm_ecef2enu(latgd, lon, alt)
        np.testing.assert_allclose(dcm_ecef2enu, dcm_ecef2enu_expected)
   
    def test_pole_prime_meridian(self):
        latgd = np.pi/2
        lon = 0
        alt = 0
        
        dcm_ecef2enu_expected = np.array([[0, 1, 0],
                                          [-1, 0, 0], 
                                          [0, 0, 1]])
        dcm_ecef2enu = transform.dcm_ecef2enu(latgd, lon, alt)
        np.testing.assert_array_almost_equal(dcm_ecef2enu, dcm_ecef2enu_expected)

    def test_so3(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2enu(latgd, lon, alt)
        np.testing.assert_allclose(np.linalg.det(dcm), 1)
        np.testing.assert_array_almost_equal(dcm.T.dot(dcm), np.eye(3,3))
    
    def test_inverse_identity(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2enu(latgd, lon, alt)
        dcm_opp = transform.dcm_enu2ecef(latgd, lon, alt)
        np.testing.assert_allclose(dcm.T, dcm_opp)

class TestECEF2SEZ():
    def test_equator_prime_meridian(self):
        latgd = 0
        lon = 0
        alt = 0
        dcm_expected = np.array([[0, 0, -1], 
                                 [0, 1, 0], 
                                 [1, 0, 0]])
        dcm = transform.dcm_ecef2sez(latgd, lon, alt)
        np.testing.assert_array_almost_equal(dcm, dcm_expected)
   
    def test_pole_prime_meridian(self):
        latgd = np.pi/2
        lon = 0
        alt = 0
        
        dcm_expected = np.array([[1, 0, 0],
                                 [0, 1, 0], 
                                 [0, 0, 1]])
        dcm = transform.dcm_ecef2sez(latgd, lon, alt)
        np.testing.assert_array_almost_equal(dcm , dcm_expected)

    def test_so3(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2sez(latgd, lon, alt)
        np.testing.assert_allclose(np.linalg.det(dcm), 1)
        np.testing.assert_array_almost_equal(dcm.T.dot(dcm), np.eye(3,3))
    
    def test_inverse_identity(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2sez(latgd, lon, alt)
        dcm_opp = transform.dcm_sez2ecef(latgd, lon, alt)
        np.testing.assert_allclose(dcm.T, dcm_opp)
    
    def test_dcm_loop(self):
        latitude = np.linspace(-np.pi/2, np.pi/2, 10)
        longitude = np.linspace(-np.pi, np.pi, 10)
        for lat, lon in zip(latitude, longitude):
            dcm_expected = rot2(np.pi/2 - lat, 'r').dot(rot3(lon, 'r'))
            dcm = transform.dcm_ecef2sez(lat, lon)
            np.testing.assert_allclose(dcm, dcm_expected)

class TestECEF2NED():


    def test_equator_prime_meridian(self):
        latgd = 0
        lon = 0
        alt = 0
        dcm_expected = np.array([[0, 0, 1], 
                                 [0, 1, 0], 
                                 [-1, 0, 0]])
        dcm = transform.dcm_ecef2ned(latgd, lon, alt)
        np.testing.assert_allclose(dcm, dcm_expected)
   
    def test_pole_prime_meridian(self):
        latgd = np.pi/2
        lon = 0
        alt = 0
        
        dcm_expected = np.array([[0, 1, 0],
                                 [-1, 0, 0], 
                                 [0, 0, 1]])
        dcm = transform.dcm_ecef2enu(latgd, lon, alt)
        np.testing.assert_array_almost_equal(dcm , dcm_expected)

    def test_so3(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2ned(latgd, lon, alt)
        np.testing.assert_allclose(np.linalg.det(dcm), 1)
        np.testing.assert_array_almost_equal(dcm.T.dot(dcm), np.eye(3,3))
    
    def test_inverse_identity(self):
        latgd = np.random.uniform(-np.pi/2, np.pi/2)
        lon = np.random.uniform(-np.pi, np.pi)
        alt = 0
        dcm = transform.dcm_ecef2ned(latgd, lon, alt)
        dcm_opp = transform.dcm_ned2ecef(latgd, lon, alt)
        np.testing.assert_allclose(dcm.T, dcm_opp)

class TestPQW2LVLH():
    nu = np.random.uniform(0, 2*np.pi)
    dcm = transform.dcm_pqw2lvlh(nu)
    dcm_opp = transform.dcm_lvlh2pqw(nu)

    def test_true_anomaly_90(self):
        nu = np.deg2rad(90)
        phat = np.array([1, 0, 0])
        R_pqw2lvlh = transform.dcm_pqw2lvlh(nu)
        R_lvlh2pqw = transform.dcm_lvlh2pqw(nu)
        rhat = np.array([0, -1, 0])
        np.testing.assert_array_almost_equal(R_pqw2lvlh.dot(phat), rhat)
        np.testing.assert_array_almost_equal(R_lvlh2pqw.dot(rhat), phat)

    def test_true_anomaly_180(self):
        nu = np.deg2rad(180)
        phat = np.array([1, 0, 0])
        R_pqw2lvlh = transform.dcm_pqw2lvlh(nu)
        R_lvlh2pqw = transform.dcm_lvlh2pqw(nu)
        rhat = np.array([-1, 0, 0])
        np.testing.assert_array_almost_equal(R_pqw2lvlh.dot(phat), rhat)
        np.testing.assert_array_almost_equal(R_lvlh2pqw.dot(rhat), phat)

    def test_so3(self):
        dcm = self.dcm
        np.testing.assert_allclose(np.linalg.det(dcm), 1)
        np.testing.assert_array_almost_equal(dcm.T.dot(dcm), np.eye(3,3))
    
    def test_inverse_identity(self):
        dcm = self.dcm
        dcm_opp = self.dcm_opp
        np.testing.assert_allclose(dcm.T, dcm_opp)

class TestECI2ECEF():
    """Test to make sure we can convert a location on the earth to the correct ECEF vector
    """

    lonexp = np.deg2rad(72.5529)
    latgdexp= np.deg2rad(34.352496)
    latgcexp = np.deg2rad(34.173429)
    altexp = 5085.22 # kilometer

    eci_exp = np.array([6524.834, 6862.875, 6448.296])
    jd_exp = 2449773.0
    _, lst = time.gstlst(jd_exp, lonexp)
    eci = geodetic.site2eci(latgdexp, altexp, lst)
    ecef = geodetic.lla2ecef(latgdexp, lonexp, altexp)
    Reci2ecef = transform.dcm_eci2ecef(jd_exp)
    eci_from_ecef = Reci2ecef.T.dot(ecef)

    def test_eci_vallado(self):
        np.testing.assert_allclose(self.eci, self.eci_exp, rtol=1e-2)

    def test_eci_from_ecef(self):
        np.testing.assert_allclose(self.eci_from_ecef, self.eci_exp, rtol=1e-2)

class TestECEF2ECI():
    """Example pg.106 in BMW
    """
    lon = np.deg2rad(-57.296)
    lat = 0
    alt = 6.378 # kilometer above equator
    date = (1970, 1, 2, 6, 0, 0)
    jd, _ = time.date2jd(date[0], date[1], date[2], date[3], date[4], date[5]) 
    gst0_exp = 1.749333
    gst_exp = attitude.normalize(9.6245, 0, 2*np.pi)
    lst_exp = attitude.normalize(8.6245, 0, 2*np.pi)

    eci_exp = np.array([-0.697*6378.137, 0.718*6378.137, 0])

    gst0 = time.gsttime0(date[0])
    gst, lst = time.gstlst(jd, lon)

    eci = geodetic.site2eci(lat, alt, lst)
    ecef = geodetic.lla2ecef(lat,lon,alt)
    Reci2ecef = transform.dcm_eci2ecef(jd)
    eci_from_ecef = Reci2ecef.T.dot(ecef)

    def test_gst0(self):
        np.testing.assert_allclose(self.gst0, self.gst0_exp, rtol=1e-4)

    def test_gst(self):
        np.testing.assert_allclose(self.gst, self.gst_exp, rtol=1e-4)

    def test_lst(self):
        np.testing.assert_allclose(self.lst, self.lst_exp, rtol=1e-3)


    def test_eci(self):
        np.testing.assert_allclose(self.eci, self.eci_exp, rtol=1e-2)

    def test_eci_from_ecef(self):
        np.testing.assert_allclose(self.eci_from_ecef, self.eci_exp, rtol=1e-2)



import numpy as np
from astro import maneuver, constants, kepler
import pdb
class TestRVFPA2OrbitEl():
    r = 4 * constants.earth.radius
    v = 4.54 # km /sec
    fpa = np.deg2rad(-40)
    a_true = 37477.21798
    p_true = 19751.017296
    ecc_true = 0.687739800
    nu_true = 4.377814485
    a, p, ecc, nu = maneuver.rvfpa2orbit_el(r, v, fpa, constants.earth.mu)

    def test_semimajor_axis(self):
        """AAE532 PS6
        """
        np.testing.assert_allclose(self.a, self.a_true)

    def test_semiparameter(self):
        """AAE532 PS6
        """
        np.testing.assert_allclose(self.p, self.p_true)
    
    def test_eccentricy(self):
        np.testing.assert_allclose(self.ecc, self.ecc_true)

    def test_true_anomaly(self):
        np.testing.assert_allclose(self.nu, self.nu_true)

class TestSingleImpulse():
    # 2017 HW5 Problem 1
    rm = 59029.6
    vm = 1.69
    fpam = np.deg2rad(27.45)
    delta_v = 1.2
    alpha = np.deg2rad(30)
    rf, vf, fpaf = maneuver.single_impulse(rm, vm, fpam, delta_v, alpha)

    rf_true = rm
    vf_true = 2.7944049
    fpaf_true = np.deg2rad(39.84879352)

    def test_position(self):
        np.testing.assert_allclose(self.rf, self.rf_true) 
    
    def test_velocity(self):
        np.testing.assert_allclose(self.vf, self.vf_true)

    def test_flight_path_angle(self):
        np.testing.assert_allclose(self.fpaf, self.fpaf_true)

def test_delta_v_solve_planar():
    v1 = 2.455
    v2 = 2.867
    fpa1 = np.deg2rad(26.03)
    fpa2 = 0
    deltav, alpha, beta = maneuver.delta_v_solve_planar(v1, v2, fpa1, fpa2)
    deltav_true = 1.2639818
    alpha_true = 1.47477394
    beta_true = np.pi - alpha
    np.testing.assert_allclose(deltav, deltav_true)
    np.testing.assert_allclose(alpha, alpha_true)
    np.testing.assert_allclose(beta, beta_true)

def test_planar_orbit_intersection():
    ecc1 = 0.75
    p1 = kepler.semilatus_rectum(4.5*constants.earth.radius, ecc1)
    a2, p2, ecc2 = kepler.perapo2aecc(2 * constants.earth.radius, 6 * constants.earth.radius)
    dargp = np.deg2rad(35)

    nu = maneuver.planar_conic_orbit_intersection(p1, p2, ecc1, ecc2, dargp)
    np.testing.assert_allclose(nu, np.deg2rad(110.342156))

def test_vel_mag():
    r = 5.04 * constants.earth.radius 
    a = 6 * constants.earth.radius
    mu = constants.earth.mu
    v_true = 3.79259
    v = maneuver.vel_mag(r, a, mu)
    np.testing.assert_allclose(v, v_true, rtol=1e-4)

def test_synodic_period():
    a1 = constants.earth.orbit_sma 
    a2 = constants.mars.orbit_sma 
    mu = constants.sun.mu 
    S_true = 67377144.56697
    S = maneuver.synodic_period(a1, a2, mu)
    np.testing.assert_allclose(S, S_true, rtol=1e-4)

class TestHohmannCircular():
    r1 = constants.earth.orbit_sma
    r2 = constants.neptune.orbit_sma
    ecc1, ecc2 = 0, 0
    nu1, nu2 = 0, np.pi
    mu = constants.sun.mu
    dv1_true, dv2_true, tof_true, phase_angle_true = 11.65413, 4.05359, 9.66154e8, np.deg2rad(113.153)
    dv1, dv2, tof, phase = maneuver.hohmann(r1, r2, ecc1, ecc2, nu1, nu2, mu)

    def test_dv1(self):
        np.testing.assert_allclose(self.dv1, self.dv1_true, rtol=1e-4)

    def test_dv2(self):
        np.testing.assert_allclose(self.dv2, self.dv2_true, rtol=1e-4)

    def test_tof(self):
        np.testing.assert_allclose(self.tof, self.tof_true, rtol=1e-4)

    def test_tof(self):
        np.testing.assert_allclose(self.phase, self.phase_angle_true, rtol=1e-4)

class TestDeltaV_VNC_Planar():
    dv_mag = 0.75
    alpha = np.deg2rad(-60)
    beta = 0
    fpa = np.deg2rad(21.8014)
    dv_vnc_true = np.array([0.3750, -0.64952, 0])
    dv_lvlh_true = np.array([-0.46379, 0.589404, 0])
    dv_vnc, dv_lvlh = maneuver.delta_v_vnc(dv_mag, alpha, beta, fpa)
    
    def test_vnc(self):
        np.testing.assert_allclose(self.dv_vnc,self.dv_vnc_true, rtol=1e-4)

    def test_lvlh(self):
        np.testing.assert_allclose(self.dv_lvlh, self.dv_lvlh_true, rtol=1e-4)



import numpy as np

def test_kernel_hashes():
    """Compare the hashes fo the kernels against some known working
    version
    """

    print("Not finished yet")
    np.testing.assert_equal(1, 1)

def test_images_hashes():
    """Compare hashes of all the NEAR imagery
    """
    print("Not implemented yet")
    np.testing.assert_equal(1, 1)
    



from astro import tle, satellite
import numpy as np
import os

cwd = os.path.realpath(os.path.dirname(__file__))

class TestISSUSAFA():
    l0 = "0 ISS (ZARYA)             "
    l1 = "1 25544U 98067A   06164.43693594  .00014277  00000-0  10780-3 0  6795"
    l2 = "2 25544  51.6455 280.1294 0004346 245.9311 226.9658 15.72751720375095"
    ifile = 'Predict.dat'
    sat = tle.get_tle(os.path.join(cwd, ifile))

    def test_validtle(self):
        np.testing.assert_equal(tle.validtle(self.l0, self.l1, self.l2), True)

    def test_line1_checksum(self):
        np.testing.assert_allclose(tle.checksum(self.l1), 5)

    def test_line2_checksum(self):
        np.testing.assert_allclose(tle.checksum(self.l2), 5)

    def test_epoch_year(self):
        np.testing.assert_allclose(self.sat[0].epoch_year, 2006)

    def test_epoch_day(self):
        np.testing.assert_allclose(self.sat[0].epoch_day, 164.43693594)

    def test_ndot2_revperdaysquared(self):
        np.testing.assert_allclose(self.sat[0].tle.ndot_over_2, 0.00014277)

    def test_inclination_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.inc, 51.64550000)

    def test_raan_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.raan, 280.12940000)

    def test_ecc_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.ecc, 0.00043460)

    def test_argp_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.argp, 245.93110000)

    def test_mean_anomaly_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.ma, 226.9658000)

    def test_mean_motion_deg(self):
        np.testing.assert_allclose(self.sat[0].tle.mean_motion, 15.72751720)

    def test_mean_motion_rad(self):
        np.testing.assert_allclose(self.sat[0].n0, 0.00114373733)

    def test_ecc(self):
        np.testing.assert_allclose(self.sat[0].ecc0, 0.00043460000)

    def test_inc_rad(self):
        np.testing.assert_allclose(self.sat[0].inc0, 0.90138401884)

    def test_raan_rad(self):
        np.testing.assert_allclose(self.sat[0].raan0, 4.88918036164)

    def test_argp_rad(self):
        np.testing.assert_allclose(self.sat[0].argp0, 4.29230742805)

    def test_ndot2_radpersecsquared(self):
        np.testing.assert_allclose(self.sat[0].ndot2, 1.20168141063e-013)

    def test_eccdot_persecond(self):
        np.testing.assert_allclose(self.sat[0].eccdot, -1.40011545218e-10)

    def test_mean_anomaly_rad(self):
        np.testing.assert_allclose(self.sat[0].mean0, 3.96130049942e0)

    def test_raandot_radpersecond(self):
        np.testing.assert_allclose(self.sat[0].raandot, -1.03554877709e-6)

    def test_argpdot_radpersecond(self):
        np.testing.assert_allclose(self.sat[0].argpdot, 7.72047261206e-7)



"""Test out the propogator module
"""

import numpy as np
from astro import propogator, constants

def test_accel_third():
    """MAE3145 HW1 Problem 3"""
    mu = constants.sun.mu
    r_moon2sun = np.array([-constants.earth.orbit_sma, -constants.moon.orbit_sma])
    r_earth2sun = np.array([-constants.earth.orbit_sma, 0])
    direct, indirect, perturbation = propogator.accel_third(mu, r_moon2sun, r_earth2sun)
    np.testing.assert_allclose(direct, (-5.930645e-6, -1.523994e-8), rtol=1e-4)
    np.testing.assert_allclose(indirect, (5.930704e-6, 0), rtol=1e-4)
    np.testing.assert_allclose(perturbation, (5.874308e-11, -1.523994e-8), rtol=1e-4)

def test_accel_twobody():
    me = constants.earth.mass
    mb = constants.moon.mass
    r = np.array([0, constants.moon.orbit_sma])
    accel = propogator.accel_twobody(me, mb, r)
    np.testing.assert_allclose(accel, (0, -2.7307e-6), rtol=1e-4)




"""test all the functions in planets 
"""
import numpy as np

from astro import planets, time
import pdb


rtol = 1e-0

def test_sun_earth_eci_usafa():
    jd = 2453905.50000000
    expected_rsun = [6371243.918400, 139340081.269385, 60407749.811252]
    rsun, ra, dec = planets.sun_earth_eci(jd)
    np.testing.assert_allclose(rsun, expected_rsun, rtol=1e-4)

def test_sun_eci_vallado():
    jd, mjd = time.date2jd(1994, 4, 2, 0, 0, 0)
    jd_vallado = 2449444.5
    rsun_vallado = np.array([146241097,
                             28574940,
                             12389196])
    ra_vallado, dec_vallado = np.deg2rad(11.056071), np.deg2rad(4.7529393)
    rsun, ra, dec = planets.sun_earth_eci(jd)


    np.testing.assert_allclose(jd, jd_vallado)
    np.testing.assert_allclose(rsun, rsun_vallado, rtol=1e-4)
    np.testing.assert_allclose((ra, dec), (ra_vallado, dec_vallado), rtol=1e-4)

class TestMercuryCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 0)
    
    eccexp = 2.056334003441381e-01
    incexp = 7.003938518641813e+00
    raanexp = 4.830876318589697e+01
    argpexp = 2.917536389390346e+01
    nuexp = 2.786392215752991e+01
    pexp = 3.870982978406551e-1 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestVenusCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 1)
    
    eccexp = 6.792032469514713e-03
    incexp = 3.394477356853931e+00
    raanexp = 7.662882872939052e+01
    argpexp = 5.473647539038130e+01
    nuexp = 1.975782148111063e+02
    pexp = 7.233262662615381e-01 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)


class TestEarthCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 2)
    
    eccexp =1.579421783270573e-02
    incexp = 3.001710480218142e-03
    raanexp = 1.482633325004871e+02
    argpexp = 3.153399510406209e+02
    nuexp = 1.688484428230512e+02
    pexp = 1.000936277866668e+00 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(-np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestMarsCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 3)
    
    eccexp = 9.352430073957714e-02
    incexp = 1.848364578438978e+00
    raanexp = 4.950693065746390e+01
    argpexp = 2.866247286888100e+02
    nuexp = 1.327364669246639e+02
    pexp = 1.523603073362533e+00 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestJupiterCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 4)
    
    eccexp = 4.886730265850301e-02
    incexp =  1.303742758808121e+00
    raanexp = 1.005110590888574e+02
    argpexp = 2.737359701611508e+02
    nuexp = 1.895852218694279e+02
    pexp = 5.202296207914233e+00 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestSaturnCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 5)
    
    eccexp = 5.196085002126025e-02
    incexp = 2.487587507263742e+00
    raanexp = 1.135815401076222e+02
    argpexp = 3.402669207862293e+02
    nuexp = 1.706484255046402e+02
    pexp = 9.568736466329314e+00 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestUranusCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 6)
    
    eccexp = 4.964052992824323e-02
    incexp = 7.701258737658703e-01
    raanexp = 7.413417028463512e+01
    argpexp = 9.896398369052795e+01
    nuexp = 2.118953182102191e+02
    pexp = 1.913027860362278e+01 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestNeptuneCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 7)
    
    eccexp = 6.564105141096746e-03
    incexp = 1.766501817898360e+00
    raanexp = 1.317099904241699e+02
    argpexp = 2.795450061842975e+02
    nuexp = 2.909187977999030e+02
    pexp = 3.001973084229905e+01 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)

class TestPlutoCOE():
    jd = 2457928.5
    coe, _, _, _, _ = planets.planet_coe(jd, 8)
    
    eccexp = 2.496793714912267e-01
    incexp = 1.711614987378614e+01
    raanexp = 1.102980595626739e+02
    argpexp = 1.121586170630204e+02
    nuexp = 6.522471039312354e+01
    pexp = 3.929274514682939e+01 * (1 - eccexp**2) 

    def test_p(self):
        np.testing.assert_allclose(self.coe.p, self.pexp, rtol=rtol)

    def test_ecc(self):
        np.testing.assert_allclose(self.coe.ecc, self.eccexp, rtol=rtol)

    def test_inc(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.inc), self.incexp, rtol=rtol)

    def test_raan(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.raan), self.raanexp, rtol=rtol)

    def test_argp(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.argp), self.argpexp, rtol=rtol)

    def test_nu(self):
        np.testing.assert_allclose(np.rad2deg(self.coe.nu), self.nuexp, rtol=rtol)



import numpy as np
from astro import constants


def test_newton_gravity_constant():
    G = 6.673e-20
    np.testing.assert_equal(constants.G, G)

def test_rad2deg():
    np.testing.assert_allclose(np.pi/2*constants.rad2deg, np.rad2deg(np.pi/2))

def test_deg2rad():
    np.testing.assert_allclose(100*constants.deg2rad, np.deg2rad(100))

def test_sec2hr():
    np.testing.assert_allclose(3600*constants.sec2hr, 1)
    np.testing.assert_allclose(1*constants.hr2sec, 3600)

def test_sec2day():
    np.testing.assert_allclose(1 * constants.sec2day , 1 / (60 * 60 * 24))
    np.testing.assert_allclose(1 * constants.day2sec, 24 * 60 * 60)




"""Satellite Module

Define satellite object from a TLE

Notes
-----

Attributes
----------
module_level_variable1 : int
    Descrption of the variable

Author
------
Shankar Kulumani		GWU		skulumani@gwu.edu
"""
import numpy as np
from collections import namedtuple
from astro import constants, time, kepler, geodetic
from kinematics import attitude
import pdb

import matplotlib.pyplot as plt
import matplotlib.lines as mlines
import matplotlib.patches as mpatches

deg2rad = np.pi / 180
rad2deg = 180 / np.pi
sec2day = 1 / (24 * 3600)
day2sec = 24 * 3600

COE = namedtuple('COE', ['n', 'ecc', 'raan', 'argp', 'mean', 'E', 'nu', 'a',
                         'p', 'inc'])

PASS = namedtuple('PASS', ['jd', 'az', 'el', 'site_eci', 'sat_eci', 'gst', 'lst',
                           'sun_alt', 'alpha', 'beta', 'sun_eci', 'rho'])

RADAR_PASS = namedtuple('RADAR_PASS', ['jd', 'site_eci', 'sat_r_eci', 'sat_v_eci', 
                                       'gst', 'lst', 'sun_eci', 'sun_alt', 'alpha', 
                                       'beta', 'rho', 'az', 'el', 'drho', 'daz', 'dele'])

class Satellite(object):
    """Satellite class storing all the necessary information for a single
    satellite extracted from a TLE from space-track.org
    """

    def __init__(self, elements):
        # now do some conversions of the TLE components
        epoch_year = (2000 + elements.epoch_year
                      if elements.epoch_year < 70
                      else 1900 + elements.epoch_year)

        # working units
        nddot6 = elements.nddot_over_6 * (2 * np.pi) * (sec2day**3)
        ndot2 = elements.ndot_over_2 * (2 * np.pi) * (sec2day**2)
        inc0 = elements.inc * deg2rad
        raan0 = elements.raan * deg2rad
        argp0 = elements.argp * deg2rad
        mean0 = elements.ma * deg2rad
        n0 = elements.mean_motion * 2 * np.pi * sec2day
        ecc0 = elements.ecc
        epoch_day = elements.epoch_day

        # calculate perturbations raan, argp, ecc
        raandot, argpdot, eccdot, adot = j2dragpert(inc0, ecc0, n0, ndot2)

        # calculate epoch julian date
        mo, day, hour, mn, sec = time.dayofyr2mdhms(epoch_year, epoch_day)
        epoch_jd, _ = time.date2jd(epoch_year, mo, day, hour, mn, sec)

        # store with class
        self.satname = elements.satname
        self.satnum = elements.satnum
        self.tle = elements
        self.ndot2 = ndot2
        self.nddot6 = nddot6
        self.inc0 = inc0
        self.raan0 = raan0
        self.ecc0 = ecc0
        self.argp0 = argp0
        self.mean0 = mean0
        self.n0 = n0
        self.epoch_jd = epoch_jd
        self.epoch_year = epoch_year
        self.epoch_day = epoch_day

        self.raandot = raandot
        self.argpdot=argpdot
        self.eccdot = eccdot
        self.adot = adot

    def tle_update(self, jd_span, mu=398600.5):
        """Update the state of satellite given a JD time span

        This procedure uses the method of general perturbations to update
        classical elements from epoch to a future time for inclined elliptical
        orbits.  It includes the effects due to first order secular rates
        (second order for mean motion) caused by drag and J2.

        Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

        Inputs:
            deltat - elapsed time since epoch (sec)
            n0 - mean motion at epoch (rad/sec)
            ndot2 - mean motion rate divided by 2 (rad/sec^2)
            ecc0 - eccentricity at epoch (unitless)
            eccdot - eccentricity rate (1/sec)
            raan0 - RAAN at epoch (rad)
            raandot - RAAN rate (rad/sec)
            argp0 - argument of periapsis at epoch (rad)
            argdot - argument of periapsis rate (rad/sec)
            mean0 - mean anomaly at epoch

        Outputs:
            n - mean motion at epoch + deltat (rad/sec)
            ecc -  eccentricity at epoch + deltat (unitless)
            raan - RAAN at epoch + deltat (rad)
            argp - argument of periapsis at epoch + deltat (rad)
            nu -  true anomaly at epoch + deltat (rad)

        Globals: None

        Constants: None

        Coupling:
            newton.m

        References:
            Astro 321 PREDICT LSN 24-25
        """

        deltat = (jd_span - self.epoch_jd) * day2sec

        # epoch orbit parameters
        n0 = self.n0
        ecc0 = self.ecc0
        ndot2 = self.ndot2
        nddot6 = self.nddot6
        raan0 = self.raan0
        argp0 = self.argp0
        mean0 = self.mean0
        inc0 = self.inc0
        adot = self.adot
        eccdot = self.eccdot
        raandot = self.raandot
        argpdot = self.argpdot

        _, nu0, _ = kepler.kepler_eq_E(mean0, ecc0)
        a0 = kepler.n2a(n0, mu)
        M0, E0 = kepler.nu2anom(nu0, ecc0)

        # propogated elements
        a1 = a0 + adot * deltat # either use this or compute a1 from n1 instead
        ecc1 = ecc0 + eccdot * deltat
        raan1 = raan0 + raandot * deltat
        argp1 = argp0 + argpdot * deltat
        inc1 = inc0 * np.ones_like(ecc1)
        n1 = n0 + ndot2 * 2 * deltat
        # a1 = kepler.n2a(n1, mu)
        p1 = kepler.semilatus_rectum(a1, ecc1)

        M1 = M0 + n0 * deltat+ ndot2 *deltat**2 + nddot6 *deltat**3
        M1 = attitude.normalize(M1, 0, 2 * np.pi)
        E1, nu1, _ = kepler.kepler_eq_E(M1, ecc1)

        # convert to vectors
        r_eci, v_eci, _, _ = kepler.coe2rv(p1, ecc1, inc1, raan1, argp1, nu1, mu)

        # save all of the variables to the object
        self.jd_span = jd_span
        self.coe = COE(n=n1, ecc=ecc1, raan=raan1, argp=argp1, mean=M1,
                       E=E1, nu=nu1, a=a1, p=p1, inc=inc1)

        self.r_eci = r_eci
        self.v_eci = v_eci

        return 0
    
    # TODO Reformat visibility check to be more efficient. Don't ahve to compute and store all sun vectors
    def visible(self, site):
        """Check if current sat is visible from the site
        """
        sat_eci = self.r_eci
        site_eci = site['eci']
        sun_eci = site['sun_eci']

        alpha = np.arccos(np.einsum('ij,ij->i', site_eci, sun_eci) /
                          np.linalg.norm(site_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                            axis=1))
        beta = np.arccos(np.einsum('ij,ij->i', sat_eci, sun_eci) /
                         np.linalg.norm(sat_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                          axis=1))
        sun_alt = np.linalg.norm(sun_eci, axis=1) * np.sin(beta)

        jd_vis = []
        rho_vis = []
        az_vis = []
        el_vis = []

        # create array to store all the passes
        all_passes = []
        current_pass = PASS(jd=[], az=[], el=[], site_eci=[], sat_eci=[],
                            gst=[], lst=[], sun_alt=[], alpha=[], beta=[],
                            sun_eci=[], rho=[])

        jd_last = self.jd_span[0]
        for (jd, a, b, sa, si, su, su_alt, lst, gst) in zip(self.jd_span, alpha,
                                                            beta, sat_eci,
                                                            site_eci, sun_eci,
                                                            sun_alt, site['lst'],
                                                            site['gst']):

            if a > np.pi / 2:
                rho, az, el = geodetic.rhoazel(sa, si, site['lat'], lst)

                if rho < 1500 and el > 10 * deg2rad:
                    if b < np.pi / 2 or su_alt > 6378.137:
                        jd_vis.append(jd)
                        rho_vis.append(rho)
                        az_vis.append(az)
                        el_vis.append(el)

                        # we have a visible state - save to current pass if
                        # time since last visible is small enough if not then
                        # we have to create another pass namedtuple
                        if np.absolute(jd_last - jd) > (10 / 24 / 60):
                            if current_pass.jd:  # new pass
                                all_passes.append(current_pass)
                                current_pass = PASS(jd=[], az=[], el=[],
                                                    site_eci=[], sat_eci=[],
                                                    gst=[], lst=[], sun_alt=[],
                                                    alpha=[], beta=[],
                                                    sun_eci=[], rho=[])

                        current_pass.jd.append(jd)
                        current_pass.az.append(az)
                        current_pass.el.append(el)
                        current_pass.site_eci.append(si)
                        current_pass.sat_eci.append(sa)
                        current_pass.gst.append(gst)
                        current_pass.lst.append(lst)
                        current_pass.sun_alt.append(su_alt)
                        current_pass.alpha.append(a)
                        current_pass.beta.append(b)
                        current_pass.sun_eci.append(su)
                        current_pass.rho.append(rho)
                        jd_last = jd

        all_passes.append(current_pass)

        self.jd_vis = jd_vis
        self.rho_vis = rho_vis
        self.az_vis = az_vis
        self.el_vis = el_vis
        self.pass_vis = all_passes

    # TODO: Update documentation
    def visible_radar(self, site):
        """Check if current sat is visible from the site given an example radar sensor
        """
        sat_r_eci = self.r_eci
        sat_v_eci = self.v_eci
        site_eci = site['eci']
        sun_eci = site['sun_eci']

        alpha = np.arccos(np.einsum('ij,ij->i', site_eci, sun_eci) /
                          np.linalg.norm(site_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                            axis=1))
        beta = np.arccos(np.einsum('ij,ij->i', sat_r_eci, sun_eci) /
                         np.linalg.norm(sat_r_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                          axis=1))
        sun_alt = np.linalg.norm(sun_eci, axis=1) * np.sin(beta)

        jd_vis = []
        rho_vis = []
        az_vis = []
        el_vis = []
        drho_vis = []
        daz_vis = []
        dele_vis = []

        # create array to store all the passes
        all_passes = []
        current_pass = RADAR_PASS(jd=[], site_eci=[], sat_r_eci=[],
                                  sat_v_eci=[], gst=[], lst=[], sun_alt=[],
                                  alpha=[], beta=[], sun_eci=[], 
                                  rho=[], az=[], el=[], drho=[], daz=[], dele=[])

        jd_last = self.jd_span[0]
        for (jd, a, b, sar, sav, si, su, su_alt, lst, gst) in zip(self.jd_span, alpha,
                                                            beta, sat_r_eci, sat_v_eci,
                                                            site_eci, sun_eci,
                                                            sun_alt, site['lst'],
                                                            site['gst']):
            # deterministic radar measurement
            rho, az, el, drho, daz, dele = geodetic.rv2rhoazel(sar, sav, site['lat'],
                                                               site['lon'], site['alt'], jd)
            if el > 10 * deg2rad:
                jd_vis.append(jd)
                rho_vis.append(rho)
                az_vis.append(az)
                el_vis.append(el)
                drho_vis.append(drho)
                daz_vis.append(daz)
                dele_vis.append(dele)

                # we have a visible state - save to current pass if
                # time since last visible is small enough if not then
                # we have to create another pass namedtuple
                if np.absolute(jd_last - jd) > (10 / 24 / 60):
                    if current_pass.jd:  # new pass
                        all_passes.append(current_pass)
                        current_pass = RADAR_PASS(jd=[], site_eci=[], sat_r_eci=[],
                                  sat_v_eci=[], gst=[], lst=[], sun_alt=[],
                                  alpha=[], beta=[], sun_eci=[], 
                                  rho=[], az=[], el=[], drho=[], daz=[], dele=[])

                current_pass.jd.append(jd)
                current_pass.site_eci.append(si)
                current_pass.sat_r_eci.append(sar)
                current_pass.sat_v_eci.append(sav)
                current_pass.gst.append(gst)
                current_pass.lst.append(lst)
                current_pass.sun_alt.append(su_alt)
                current_pass.alpha.append(a)
                current_pass.beta.append(b)
                current_pass.sun_eci.append(su)
                current_pass.rho.append(rho)
                current_pass.az.append(az)
                current_pass.el.append(el)
                current_pass.drho.append(drho)
                current_pass.daz.append(daz)
                current_pass.dele.append(dele)
                jd_last = jd

        all_passes.append(current_pass)

        self.jd_vis = jd_vis
        self.rho_vis = rho_vis
        self.az_vis = az_vis
        self.el_vis = el_vis
        self.drho_vis = drho_vis
        self.daz_vis = daz_vis
        self.dele_vis = dele_vis
        self.pass_vis = all_passes
    

    def output(self, filename):
        """Write to output file
        """
        space = '    '
        with open(filename, 'a') as f:

            f.write('%s %05.0f\n' % (self.satname, self.satnum))
            f.write(
                'PASS    MON/DAY    HR:MIN(UT)    RHO(KM)    AZ(DEG)    EL(DEG)    SAT          \n')
            f.write(
                '-------------------------------------------------------------------------------\n\n')

            for ii, cur_pass in enumerate(self.pass_vis):
                for jd, rho, az, el in zip(cur_pass.jd, cur_pass.rho, cur_pass.az, cur_pass.el):
                    # convert julian day to normal date
                    yr, mo, day, hr, mn, sec = time.jd2date(jd)

                    # if sec > 30:
                    #     mn = mn + 1
                    #     if mn == 60:
                    #         hr = hr + 1
                    #         mn = 0
                    f.write('%4.0f%s' % (ii, space))
                    f.write('%3.0f/%3.0f%s' % (mo, day, space))
                    f.write('%02.0f:%02.0f:%3.1f%s' % (hr, mn, sec, space))
                    f.write('%7.2f%s' % (rho, space))
                    f.write('%7.2f%s' % (az * 180 / np.pi, space))
                    f.write('%7.2f%s' % (el * 180 / np.pi, space))
                    f.write('%13s\n' % (self.satname))

    def plot_pass(self, pass_num):
        """Try and plot a pass on a polar plot
        """
        def mapr(r):
            return 80 - r

        jd = self.pass_vis[pass_num].jd
        az = self.pass_vis[pass_num].az
        el = self.pass_vis[pass_num].el

        sy, smo, sd, sh, smn, ss = time.jd2date(jd[0])
        ey, emo, ed, eh, emn, es = time.jd2date(jd[-1])

        start_string = '{:02.0f}:{:02.0f}:{:02.0f}'.format(sh, smn, ss)
        end_string = '{:02.0f}:{:02.0f}:{:02.0f}'.format(eh, emn, es)

        fig = plt.figure()
        ax = plt.subplot(111, projection='polar')

        line = ax.plot(az, mapr(np.rad2deg(el)))[0]
        
        ax.plot(az[0], mapr(np.rad2deg(el[0])), marker='.', color='green', markersize=25)
        ax.plot(az[-1], mapr(np.rad2deg(el[-1])), marker='.', color='red', markersize=25)

        ax.set_yticks(range(-10, 90, 10))
        ax.set_yticklabels(map(str, range(90, -10, -10)))
        ax.set_theta_zero_location("N")
        fig.suptitle("%s" %
                     (self.satname), y=1.05)
        plt.title('%2d/%2d' % (smo, sd))
        plt.title('Start:\n' + start_string, loc='left')
        plt.title('End: \n' + end_string, loc='right')

        plt.show()

def j2dragpert(inc0, ecc0, n0, ndot2, mu=398600.5, re=6378.137, J2=0.00108263):
    """Perturbed Kepler Propogator including J2 and drag

    Inputs:
    inc0 - initial inclination (radians)
    ecc0 - initial eccentricity (unitless)
    n0 - initial mean motion (rad/sec)
    ndot2 - mean motion rate divided by 2 (rad/sec^2)

    Outputs:
    raandot - nodal rate (rad/sec)
    argdot - argument of periapsis rate (rad/sec)
    eccdot - eccentricity rate (1/sec)
    adot - semi major axis rate (km/sec)

    Globals: None

    Constants:
    mu - 398600.5 Earth gravitational parameter in km^3/sec^2
    re - 6378.137 Earth radius in km
    J2 - 0.00108263 J2 perturbation factor

    Coupling: 
        kepler.semilatus_rectum - compute semilatus rectum

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741
            12 5 2014 - modified to remove global varaibles
            17 Jun 2017 - Now in Python for awesomeness
            29 Nov 2017 - Include drag effect on a, ecc

    References:
        Vallado
    """

    # calculate initial semi major axis and semilatus rectum
    a0 = (mu / n0**2) ** (1 / 3)
    p0 = kepler.semilatus_rectum(a0, ecc0)

    # mean motion with J2
    nvec = n0 * (1 + 1.5 * J2 * (re / p0)**2 *
                 np.sqrt(1 - ecc0**2) * (1 - 1.5 * np.sin(inc0)**2))

    # eccentricity rate
    eccdot = (-2 * (1 - ecc0) * 2 * ndot2) / (3 * nvec)

    # calculate nodal rate
    raandot = (-1.5 * J2 * (re / p0)**2 * np.cos(inc0)) * nvec

    # argument of periapsis rate
    argdot = (1.5 * J2 * (re / p0)**2 * (2 - 2.5 * np.sin(inc0)**2)) * nvec
    
    # semi major axis rate
    adot = - 2 / 3 * a0 / nvec * 2 * ndot2

    return (raandot, argdot, eccdot, adot)

def tle_update(sat, jd_span, mu=398600.5):
    """Update the state of satellite given a JD time span

    This procedure uses the method of general perturbations to update
    classical elements from epoch to a future time for inclined elliptical
    orbits.  It includes the effects due to first order secular rates
    (second order for mean motion) caused by drag and J2.

    Author:   C2C Shankar Kulumani   USAFA/CS-19   719-333-4741

    Inputs:
        deltat - elapsed time since epoch (sec)
        n0 - mean motion at epoch (rad/sec)
        ndot2 - mean motion rate divided by 2 (rad/sec^2)
        ecc0 - eccentricity at epoch (unitless)
        eccdot - eccentricity rate (1/sec)
        raan0 - RAAN at epoch (rad)
        raandot - RAAN rate (rad/sec)
        argp0 - argument of periapsis at epoch (rad)
        argdot - argument of periapsis rate (rad/sec)
        mean0 - mean anomaly at epoch

    Outputs:
        n - mean motion at epoch + deltat (rad/sec)
        ecc -  eccentricity at epoch + deltat (unitless)
        raan - RAAN at epoch + deltat (rad)
        argp - argument of periapsis at epoch + deltat (rad)
        nu -  true anomaly at epoch + deltat (rad)

    Globals: None

    Constants: None

    Coupling:
        newton.m

    References:
        Astro 321 PREDICT LSN 24-25
    """

    deltat = (jd_span - sat.epoch_jd) * day2sec

    # epoch orbit parameters
    n0 = sat.n0
    ecc0 = sat.ecc0
    ndot2 = sat.ndot2
    nddot6 = sat.nddot6
    raan0 = sat.raan0
    argp0 = sat.argp0
    mean0 = sat.mean0
    inc0 = sat.inc0
    adot = sat.adot
    eccdot = sat.eccdot
    raandot = sat.raandot
    argpdot = sat.argpdot

    _, nu0, _ = kepler.kepler_eq_E(mean0, ecc0)
    a0 = kepler.n2a(n0, mu)
    M0, E0 = kepler.nu2anom(nu0, ecc0)

    # propogated elements
    a1 = a0 + adot * deltat # either use this or compute a1 from n1 instead
    ecc1 = ecc0 + eccdot * deltat
    raan1 = raan0 + raandot * deltat
    argp1 = argp0 + argpdot * deltat
    inc1 = inc0 * np.ones_like(ecc1)
    n1 = n0 + ndot2 * 2 * deltat
    # a1 = kepler.n2a(n1, mu)
    p1 = kepler.semilatus_rectum(a1, ecc1)

    M1 = M0 + n0 * deltat+ ndot2 *deltat**2 + nddot6 *deltat**3
    M1 = attitude.normalize(M1, 0, 2 * np.pi)
    E1, nu1, _ = kepler.kepler_eq_E(M1, ecc1)

    # convert to vectors
    r_eci, v_eci, _, _ = kepler.coe2rv(p1, ecc1, inc1, raan1, argp1, nu1, mu)

    return r_eci

def visible(sat_eci, site, jd_span):
    """Check if current sat is visible from the site
    """
    site_eci = site['eci']
    sun_eci = site['sun_eci']

    alpha = np.arccos(np.einsum('ij,ij->i', site_eci, sun_eci) /
                        np.linalg.norm(site_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                        axis=1))
    beta = np.arccos(np.einsum('ij,ij->i', sat_eci, sun_eci) /
                        np.linalg.norm(sat_eci, axis=1) / np.linalg.norm(sun_eci,
                                                                        axis=1))
    sun_alt = np.linalg.norm(sun_eci, axis=1) * np.sin(beta)

    jd_vis = []
    rho_vis = []
    az_vis = []
    el_vis = []

    # create array to store all the passes
    all_passes = []
    current_pass = PASS(jd=[], az=[], el=[], site_eci=[], sat_eci=[],
                        gst=[], lst=[], sun_alt=[], alpha=[], beta=[],
                        sun_eci=[], rho=[])

    jd_last = jd_span[0]
    for (jd, a, b, sa, si, su, su_alt, lst, gst) in zip(jd_span, alpha,
                                                        beta, sat_eci,
                                                        site_eci, sun_eci,
                                                        sun_alt, site['lst'],
                                                        site['gst']):

        if a > np.pi / 2:
            rho, az, el = geodetic.rhoazel(sa, si, site['lat'], lst)

            if rho < 1500 and el > 10 * deg2rad:
                if b < np.pi / 2 or su_alt > 6378.137:
                    jd_vis.append(jd)
                    rho_vis.append(rho)
                    az_vis.append(az)
                    el_vis.append(el)

                    # we have a visible state - save to current pass if
                    # time since last visible is small enough if not then
                    # we have to create another pass namedtuple
                    if np.absolute(jd_last - jd) > (10 / 24 / 60):
                        if current_pass.jd:  # new pass
                            all_passes.append(current_pass)
                            current_pass = PASS(jd=[], az=[], el=[],
                                                site_eci=[], sat_eci=[],
                                                gst=[], lst=[], sun_alt=[],
                                                alpha=[], beta=[],
                                                sun_eci=[], rho=[])

                    current_pass.jd.append(jd)
                    current_pass.az.append(az)
                    current_pass.el.append(el)
                    current_pass.site_eci.append(si)
                    current_pass.sat_eci.append(sa)
                    current_pass.gst.append(gst)
                    current_pass.lst.append(lst)
                    current_pass.sun_alt.append(su_alt)
                    current_pass.alpha.append(a)
                    current_pass.beta.append(b)
                    current_pass.sun_eci.append(su)
                    current_pass.rho.append(rho)
                    jd_last = jd

    all_passes.append(current_pass)
    return all_passes

def parallel_predict(sat, jd_span, site):
    eci = tle_update(sat, jd_span)
    all_passes = visible(eci, site, jd_span)

    return all_passes

def output(sat, pass_vis, filename):
    """Write to output file
    """
    space = '    '
    with open(filename, 'a') as f:

        f.write('%s %05.0f\n' % (sat.satname, sat.satnum))
        f.write(
            'PASS    MON/DAY    HR:MIN(UT)    RHO(KM)    AZ(DEG)    EL(DEG)    SAT          \n')
        f.write(
            '-------------------------------------------------------------------------------\n\n')

        for ii, cur_pass in enumerate(pass_vis):
            for jd, rho, az, el in zip(cur_pass.jd, cur_pass.rho, cur_pass.az, cur_pass.el):
                # convert julian day to normal date
                yr, mo, day, hr, mn, sec = time.jd2date(jd)

                # if sec > 30:
                #     mn = mn + 1
                #     if mn == 60:
                #         hr = hr + 1
                #         mn = 0
                f.write('%4.0f%s' % (ii, space))
                f.write('%3.0f/%3.0f%s' % (mo, day, space))
                f.write('%02.0f:%02.0f:%3.1f%s' % (hr, mn, sec, space))
                f.write('%7.2f%s' % (rho, space))
                f.write('%7.2f%s' % (az * 180 / np.pi, space))
                f.write('%7.2f%s' % (el * 180 / np.pi, space))
                f.write('%13s\n' % (sat.satname))



"""Test SPICEYPY
"""

from __future__ import absolute_import, division, print_function, unicode_literals
import spiceypy as spice
import numpy as np
import sys
import os
import pdb

from astro import kernels
import pytest
# pytestmark = pytest.mark.skip('WIP')

def test_spiceypy_installation_correct():
    spice_version = 'CSPICE_N0066'
    np.testing.assert_equal(spice.tkvrsn('TOOLKIT'),spice_version)

class TestSpiceyPyFunctions():
    cass = kernels.CassiniKernels()
    spice.furnsh(cass.metakernel) 
    utc = ['Jun 20, 2004', 'Dec 1, 2005']

    etOne = spice.str2et(utc[0])
    etTwo = spice.str2et(utc[1])
    step = 4000
    times = np.linspace(etOne, etTwo, step) 
    spice.kclear()
    def test_spiceypy_cassini(self):
        spice.furnsh(self.cass.metakernel)
        true_initial_pos = [-5461446.61080924 ,-4434793.40785864 ,-1200385.93315424]
        positions, lightTimes = spice.spkpos('Cassini', self.times, 'J2000', 'None', 'SATURN BARYCENTER')
        np.testing.assert_array_almost_equal(positions[0], true_initial_pos)
        spice.kclear()

    def test_spicepy_rotation_matrix_identity(self):
        spice.furnsh(self.cass.metakernel)
        R_E2E = spice.pxform('IAU_EARTH', 'IAU_EARTH', self.etOne)
        np.testing.assert_array_almost_equal(R_E2E, np.eye(3))
        spice.kclear()

    # @pytest.mark.skip('failing')
    def test_spicepy_state_transformation(self):
        spice.furnsh(self.cass.metakernel)
        T = spice.sxform('IAU_EARTH', 'IAU_SATURN', self.etOne)
        R = spice.pxform('IAU_EARTH', 'IAU_SATURN', self.etOne)
        (Rout, wout) = spice.xf2rav(T)
        np.testing.assert_array_almost_equal(Rout, R)
        spice.kclear()

class TestNEARKernels():
    near = kernels.NearKernels()
    
    def test_msi_frame_transformation(self):
        """Check the frame at a specific time
        """
        spice.kclear()
        spice.furnsh(self.near.metakernel)

        utc = ['Feb 10, 2001 12:00:00 UTC']
        et = spice.str2et(utc)
        
        R_b2c = spice.pxform(self.near.near_body_frame,
                self.near.near_msi_frame, et)
        R_b2c_actual = np.array([
            [.9999988429, -.0004838414, .0014422523],
            [.0004838419, .9999998829, .0000000000],
            [-.0014422522, .0000006978, .9999989600]])
        spice.kclear()
        np.testing.assert_array_almost_equal(R_b2c, R_b2c_actual)

    def test_msi_frame_transformation_fixed(self):
        spice.kclear()
        spice.furnsh(self.near.metakernel)
        etone = spice.str2et('Jan 1, 2001 10:04:02.2 UTC')
        ettwo = spice.str2et('Feb 12, 2001 18:03:02.6 UTC')

        Rone = spice.pxform(self.near.near_body_frame, 
                self.near.near_msi_frame, etone)
        Rtwo = spice.pxform(self.near.near_body_frame,
                self.near.near_msi_frame, ettwo)
        np.testing.assert_array_almost_equal(Rone, Rtwo)
        spice.kclear()

    def test_msi_frame_transpose(self):
        spice.kclear()
        spice.furnsh(self.near.metakernel)
        utc = 'Feb 2, 2001 14:20:10 UTC'
        et = spice.str2et(utc)
        Rb2c = spice.pxform(self.near.near_body_frame, 
                self.near.near_msi_frame, et)
        Rc2b = spice.pxform(self.near.near_msi_frame,
                self.near.near_body_frame, et)
        np.testing.assert_array_almost_equal_nulp(Rb2c, Rc2b.T)
        spice.kclear()

    def test_msi_boresight_vector(self):
        spice.kclear()
        spice.furnsh(self.near.metakernel)
        bs_near_body = np.array([.9999988429, -.0004838414, .0014422523])
        bs_msi_frame = np.array([1, 0, 0])
        et = spice.str2et('Feb 12, 2001 UTC')
        Rc2b = spice.pxform(self.near.near_msi_frame,
                self.near.near_body_frame, et)
        np.testing.assert_array_almost_equal(Rc2b.dot(bs_msi_frame),
                bs_near_body)
        spice.kclear()

    def test_near_landing_coverage(self):
        spice.furnsh(self.near.metakernel)
        utc = ['Feb 12, 2001 12:00:00 UTC', 'Feb 12, 2001 20:05:00 UTC']
        etone = 35251264.18507089
        ettwo = 35280364.18507829
        np.testing.assert_almost_equal(spice.str2et(utc[0]), etone)
        np.testing.assert_almost_equal(spice.str2et(utc[1]), ettwo)
        spice.kclear()

    def test_near_body_frames(self):
        """Transformation from Body fixed frame to prime frame

        There is a constant rotation of 135 deg about the Z/Third axis
        """
        spice.furnsh(self.near.metakernel)

        ckid = spice.ckobj(self.near.Ck)[0]
        cover = spice.ckcov(self.near.Ck, ckid, False, 'INTERVAL', 0.0, 'SCLK') 
        R, av, clkout = spice.ckgpav(ckid, cover[0], 0, 'NEAR_SC_BUS')
        ang = 135*np.pi/180
        R_act = np.array([[np.cos(ang), -np.sin(ang), 0], 
            [np.sin(ang), np.cos(ang), 0],
            [0, 0, 1]])

        np.testing.assert_array_almost_equal(R, R_act)
        np.testing.assert_array_almost_equal(av, np.zeros(3))
        np.testing.assert_almost_equal(clkout, cover[0])
        spice.kclear()

    def test_earth_spk_coverage(self):
        spice.furnsh(self.near.metakernel)
        spkids = spice.spkobj(self.near.SpkPlanet)
        cover = spice.stypes.SPICEDOUBLE_CELL(1000)
        spice.spkcov(self.near.SpkPlanet, 399, cover)
        result = [x for x in cover]
        expected_result = [-633873600.0, 347630400.0] 
        np.testing.assert_array_almost_equal(result, expected_result)
        spice.kclear()

    def test_eros_spk1_coverage(self):
        spice.furnsh(self.near.metakernel)
        code = spice.bodn2c('EROS')
        cover = spice.stypes.SPICEDOUBLE_CELL(1000)
        spice.spkcov(self.near.SpkEros, code, cover)
        result = [x for x in cover]
        expected_result = [-126273600.0, 37886400.0]
        np.testing.assert_array_almost_equal(result, expected_result)
        spice.kclear()

    def test_eros_spk2_coverage(self):
        code = spice.bodn2c('EROS')
        cover = spice.stypes.SPICEDOUBLE_CELL(1000)
        spice.spkcov(self.near.SpkEros2, code, cover)
        result = [x for x in cover]
        expected_result = [-31334400.0, 81432000.0] 
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_near_spk_landed_coverage(self):
        code = spice.bodn2c('NEAR')
        cover = spice.stypes.SPICEDOUBLE_CELL(1000)
        spice.spkcov(self.near.SpkNearLanded, code, cover)
        result = [x for x in cover]
        expected_result = [35279120.0, 315576000.0]
        np.testing.assert_array_almost_equal(result, expected_result)

    def test_near_spk_orbit_coverage(self):
        code = spice.bodn2c('NEAR')
        cover = spice.stypes.SPICEDOUBLE_CELL(1000)
        spice.spkcov(self.near.SpkNearOrbit, code, cover)
        result = [x for x in cover]
        expected_result = [-43200.0, 35279120.0] 
        np.testing.assert_array_almost_equal(result, expected_result)



#!/usr/bin/env python3

"""Download and handle TLEs
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import numpy as np
from collections import namedtuple
import urllib.request

from astro import time, kepler, geodetic
from astro.satellite import Satellite
from kinematics import attitude

import pdb, argparse, datetime, getpass


# tuple to hold all the items from a single TLE
TLE = namedtuple('TLE', [
    'satname', 'satnum', 'classification', 'id_year', 'id_launch',
    'id_piece', 'epoch_year', 'epoch_day', 'ndot_over_2', 'nddot_over_6',
    'bstar', 'ephtype', 'elnum', 'checksum1',
    'inc', 'raan', 'ecc', 'argp', 'ma', 'mean_motion', 'epoch_rev',
    'checksum2', 'good'])


def get_tle_visible(filename='/tmp/visible_tle.txt'):
    """Download visible tle list from Celestrak
    """
    url = 'https://www.celestrak.com/NORAD/elements/visual.txt'
    with urllib.request.urlopen(url) as response, open(filename, 'wb') as out_file:
        data = response.read() # a `bytes` object
        out_file.write(data)

def get_tle_spacetrack(filename, flag='all'):
    r"""Download TLEs from spacetrack.org

    This function will download some TLEs from Space - track.org. It requires a
    valid user account on the website. It uses favorites lists which can be
    setup in your own account.

    Parameters
    ----------
    filename : string
        Filename to store the downloaded TLEs
    flag : string
        This string can take one of three values. It allows you decide which
        TLE set to download from the website.
        all - Download bulk catalog
        visible - Download the visible satellite list
        stations - Download personal list called STATIONS
        brightest - download personal list called BRIGHTEST
        rv2coe - list used in RV2COE
        comfix - list used in COMFIX
        propogate - list used in PROPOGATE

    Returns
    -------
    None : writes to filename

    References
    ----------

    .. [1] spacetrack - https://github.com/python-astrodynamics/spacetrack
    """
    from spacetrack import SpaceTrackClient

    password = getpass.getpass('Enter SpaceTrack.org password: ')
    st = SpaceTrackClient('shanks.k', password)
    with open(filename, 'w') as f:
        if flag == 'all':
            all_tles = st.tle_latest(ordinal=1, format='3le')
        elif flag == 'stations':
            all_tles = st.tle_latest(
                favorites='STATIONS', ordinal=1, format='3le')
        elif flag == 'brightest':
            all_tles = st.tle_latest(
                favorites='BRIGHTEST', ordinal=1, format='3le')
        elif flag == 'visible':
            all_tles = st.tle_latest(
                favorites='Visible', ordinal=1, format='3le')
        elif flag == 'rv2coe':
            all_tles = st.tle_latest(favorites='RV2COE', ordinal=1,
                                     format='3le')
        elif flag == 'comfix':
            all_tles = st.tle_latest(favorites='COMFIX', ordinal=1,
                                     format='3le')
        elif flag == 'propagate':
            all_tles = st.tle_latest(favorites='PROPAGATE', ordinal=1,
                                     format='3le')
        else:
            print("Incorrect TLE favorites flag")
            all_tles = "Incorrect flag"

        f.write(all_tles)
        return 0

def validtle(l0, l1, l2):
    r"""Ensure 3 line TLEs are valid.

    This function will check each line of a 3 line TLE and ensure they are
    valid. It assumes the format is that which is found on space-track.org.

    Parameters
    ----------
    l0 : string
        First line of TLE. This line holds the object name.
        Sometimes is starts with a 0 and other times it does not.
        Both will be valid in this code.
    l1: string
        Second line of TLE. This is first real line with real data.
        Should begin with a 1 and pass a checksum. There is also a line length
        requirement.
    l2: string
        Third line of TLE. This also contains data. Begins with a 2 and has a
        checksum and line lenght requirement.

    Returns
    -------
    boolean:
        This returns a boolean object which is the concatenation of all the
        requirements for the checking.  If this output is True then the three
        lines of the inputted TLE are probably valid.

    See Also
    --------
    checksum: another related function to calculate the line checksum
    parsetle: this will take the three lines and extract out the TLE data
    stringScientificNotationToFloat: Converts the TLE number format to float

    Notes
    -----
    TLEs are an old format, but they are well documented. In addition, the TLE
    has some inherent limitations on the expected accuracy of the ephemerides
    and the data contained within it.
    
    Author
    ------
    Shankar Kulumani 

    References
    ----------

    .. [1] VALLADO, David A. Fundamentals of Astrodynamics and Applications. 3
    ed. Microcosm Press, 2007.
    .. [2] http://celestrak.com/NORAD/documentation/tle-fmt.asp

    """
    l1_valid_start = int(l1[0]) == 1
    l2_valid_start = int(l2[0]) == 2

    l1_valid_length = len(l1) == 69
    l2_valid_length = len(l2) == 69

    l1_valid_checksum = int(l1[-1]) == checksum(l1)
    l2_valid_checksum = int(l2[-1]) == checksum(l2)
    return (l1_valid_start and l2_valid_start and
            l1_valid_length and l2_valid_length and
            l1_valid_checksum and l1_valid_checksum and l2_valid_checksum)


def checksum(line):
    r"""Compute checksum for a TLE line

    sum = checksum(line)

    Parameters
    ----------
    line : string
        A single line (string) from a TLE

    Returns
    -------
    sum : int
        The checksums for each line are calculated by adding the all numerical
        digits on that line, including the line number. One is added to the
        checksum for each negative sign (-) on that line. All other non-digit
        characters are ignored.  @note this excludes last char for the checksum
        thats already there.

    Note
    ----
    Input the line as is, no need to convert to float or anything

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu

    """
    L = line.strip()
    cksum = 0
    for i in range(68):
        c = L[i]
        if c <= '9' and c >= '0':
            cksum = cksum + int(c)
        elif c == '-':
            cksum = cksum + 1
        else:
            continue

    cksum = cksum % 10

    return cksum

def stringScientificNotationToFloat(sn):
    r"""TLE Scientific notation converter

    num = stringScientificNotationToFloat(sn)

    Parameters
    ----------
    sn : string
        String from a TLE 

    Returns
    -------
    num : float
        Output the string as a float

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu

    References
    ----------
    Specific format is 5 digits, a + or -, and 1 digit, ex: 01234-5 which is
    0.01234e-5

    Examples
    --------
    An example of how to use the function

    >>> sn = '012345-5'
    >>> stringScientificNotationToFloat(sn)
    0.01234e-5
    """ 

    return 1e-5 * float(sn[:6]) * 10**int(sn[6:])


def parsetle(l0, l1, l2):
    """Parses the three lines of a TLE to their appropriate values.

    Inputs
    ------
        l0 - first line consiting of the name of TLE 
        l1 - second line
        l2 - third line

    Outputs
    -------
    namedtuple TLE: A tuple holding all of the values associated with the TLE
        satname - satellite name
        satnum - satellite number found (0 if the requested sat was not found)
        classfication  - classification (classified, unclassified)
        id_year - launch year (last 2 digits)
        id_launch - launch number of the year
        id_piece - piece of launch (a string 3 characters long)
        epoch_year - epoch year (last 2 digits)
        epoch_day - epoch day of year
        ndot_over_2 - first time derive of mean mot. divided by 2 (rev/day^2)
        nddot_over_6 - 2nd deriv of mean mot div by 6 (rev/day^3)
        bstar - bstar drag term
        ephtype- ephemeris type
        elnum  - element number (modulo 1000)
        checksum1 - checksum of first line

        inc    - inclination (deg)
        raan   - right ascension of asc. node (deg)
        ecc    - eccentricity
        argp    - argument of perigee (deg)
        ma     - mean anomaly (deg)
        mean_motion - mean motion (revs per day)
        epoch_rev - revolution number(modulo 100,000)
        checksum2 - checksum of second line

    References
    ----------
    "Fundamental of Astrodynamics and Applictions, 2nd Ed", by Vallado
    www.celestrak.com/NORAD/documentation/tle-fmt.asp
    manpages.debian.net/cgi-bin/...
    display_man.cgi?id=e2095ebad4eb81b943fcdd55ec9b7521&format=html
    """
    try:
        # parse line zero
        satname = l0[0:23]

        # parse line one
        satnum = int(l1[2:7])
        classification = l1[7:8]
        id_year = int(l1[9:11])
        id_launch = int(l1[11:14])
        id_piece = l1[14:17]
        epoch_year = int(l1[18:20])
        epoch_day = float(l1[20:32])
        ndot_over_2 = float(l1[33:43])
        nddot_over_6 = stringScientificNotationToFloat(l1[44:52])
        bstar = stringScientificNotationToFloat(l1[53:61])
        ephtype = int(l1[62:63])
        elnum = int(l1[64:68])
        checksum1 = int(l1[68:69])

        # parse line 2
        # satellite        = int(line2[2:7])
        inc = float(l2[8:16])
        raan = float(l2[17:25])
        ecc = float(l2[26:33]) * 0.0000001
        argp = float(l2[34:42])
        ma = float(l2[43:51])
        mean_motion = float(l2[52:63])
        epoch_rev = float(l2[63:68])
        checksum2 = float(l2[68:69])

        good = True
    except: # We have a bad TLE but we're not sure where
        # parse line zero
        satname = l0[0:23]

        # parse line one
        satnum = 0
        classification = 'U'
        id_year = 0
        id_launch = 0
        id_piece = 'A'
        epoch_year = 0 
        epoch_day = 0
        ndot_over_2 = 0
        nddot_over_6 = 0
        bstar = 0
        ephtype = 0
        elnum = 0
        checksum1 = 0

        # parse line 2
        # satellite        = int(line2[2:7])
        inc = 0
        raan = 0
        ecc = 0
        argp = 0
        ma = 0
        mean_motion = 0
        epoch_rev = 0
        checksum2 = 0

        good = False

    return TLE(satnum=satnum, classification=classification, id_year=id_year,
               id_launch=id_launch, id_piece=id_piece, epoch_year=epoch_year,
               epoch_day=epoch_day, ndot_over_2=ndot_over_2,
               nddot_over_6=nddot_over_6, bstar=bstar, ephtype=ephtype,
               elnum=elnum, checksum1=checksum1, inc=inc, raan=raan, ecc=ecc,
               argp=argp, ma=ma, mean_motion=mean_motion, epoch_rev=epoch_rev,
               checksum2=checksum2, satname=satname, good=good)




def get_tle(filename):
    """Assuming a file with 3 Line TLEs is given this will parse the file
    and save all the elements to a list or something.
    """
    sats = []
    tles = 0
    lines = 0
    with open(filename, 'r') as f:
        l0 = f.readline().strip()
        while l0:
            l1 = f.readline().strip()
            lines += 1
            l2 = f.readline().strip()
            lines += 1
            # error check to make sure we have read a complete TLE
            if validtle(l0, l1, l2):
                # now we parse the tle
                elements = parsetle(l0, l1, l2)
                tles += 1
                # final check to see if we got a good TLE
                if elements.good:
                    # instantiate a bunch of instances of a Satellite class
                    sats.append(Satellite(elements))

            else:
                print("Invalid TLE")

            l0 = f.readline().strip()
            lines += 1

    print("Read {} lines which should be {} TLEs".format(lines, lines / 3))
    print("Found {} TLEs".format(tles))
    print("Parsed {} SATs".format(len(sats)))
    return sats

if __name__ == '__main__':
    # parse arguments to download tles to a file if desired
    output_name = datetime.datetime.now().isoformat() + '_tle.txt' 
    parser = argparse.ArgumentParser(description='TLE downloader')
    parser.add_argument('--output', '-o', help='Output path to save the TLEs',
                        default=output_name, action='store', type=str)
    parser.add_argument('list', help='TLE list to download', choices=['visible', 'stations', 'all', 'brightest'],
                        type=str)
    args = parser.parse_args()

    tle_flag = args.list
    ofile = args.output
    
    get_tle_spacetrack(ofile, tle_flag)
    



"""This module will download kernels and setup SPICE
"""
from __future__ import absolute_import, division, print_function, unicode_literals
import os
import time
from urllib.request import urlretrieve
import spiceypy as spice

import logging

# TODO Reduce the total amount of data that needs to be downloaded for testing spiceypy
# TODO Logging for downloading of kernels
# TODO Switch to hosting kernels on Github instead of downloading directly
# TODO Better handling of errors in downloading and verifying that kernels exist

# all kernels are saved to the astro package path
cwd = os.path.realpath(os.path.dirname(__file__))
directory = 'kernels'

if not os.path.isdir(os.path.join(cwd, directory)):
    os.mkdir(os.path.join(cwd, directory))


def getKernelNameFromUrl(url):
    r"""Parse URL to get the kernel name

    kernel_name = getKernelNameFromUrl(url)

    Parameters
    ----------
    url : str
        URL for kernel (direct link)

    Returns
    -------
    kernel_name : str
        Outputs the last part of the URL - the filename of the kernel

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """ 
    return str(url.split('/')[-1])


def getPathfromUrl(url):
    r"""Create the local path given a URL

    path = getPathfromUrl(url)

    Parameters
    ----------
    url : str
        URL location for the kernel file

    Returns
    -------
    path : str
        local path for kernel

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """
    return str(os.path.join(cwd, directory, getKernelNameFromUrl(url)))


def delete_file(path, logger=logging.getLogger(__name__)):
    r"""Delete a given kernel

    delete_file(path)

    Parameters
    ----------
    path : str
        Path to the kernel. It will be deleted

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu
    """

    if os.path.exists(path):
        os.remove(path)
        logger.info('Deleted {}'.format(path))
    else:
        logger.info('{} does not exist'.format(path))


# TODO : Create a list for each type of kernel, a list for SPK, FK, BSP, etc.
# Then have logic setup to automatically download each specific type of kernel for the object
class NearKernels(object):
    r"""SPICE Kernels for NEAR mission

    NearKernels(path, logger)

    Parameters
    ----------
    path : str
        Path to store all kernels. Defaults to CWD of astro package
    logger: logging logger object
        Logger for writing logs. If not given will be instantiated

    Notes
    -----
    This only downloads data for 2001, not the whole mission.

    Author
    ------
    Shankar Kulumani		GWU		skulumani@gwu.edu

    References
    ----------
    More data on Near is available:
    https://pdssbn.astro.umd.edu/data_sb/missions/near/index.shtml
    """

    def __init__(self, path=cwd, logger=logging.getLogger(name=__name__)):
        r"""Instantiate the NearKernel object

        Notes
        -----
        This holds all the links and paths to the NEAR kernels.
        Also will create a metakernel which can be loaded by SPICE

        Author
        ------
        Shankar Kulumani		GWU		skulumani@gwu.edu

        """
        self.logger = logger
        self.logger.debug('Instantiating NearKernels')

        self.near_id = '-93'
        self.eros_id = '2000433'
        
        self.near_body_frame = 'NEAR_SC_BUS_PRIME'
        self.near_body_frame_id = -93000
        self.near_msi_frame = 'NEAR_MSI'
        self.near_msi_frame_id = -93001
        self.eros_body_frame = 'IAU_EROS'
        self.eros_body_frame_id = 2000433

        self.inertial_frame = 'J2000'

        # mission start from near_171.tsc in UTC
        self.start_et = -122138129.0388301
        self.start_utc = '1996-02-17 20:43:28.775 UTC'

        self.Lsk_url = 'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/a_old_versions/naif0008.tls'
        self.Ck_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ck/near_20010101_20010228_v01.bc'
        self.Sclk_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/sclk/near_171.tsc'

        self.PckEros1_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/erosatt_1998329_2001157_v01.bpc'
        self.PckEros2_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/erosatt_1999304_2001151.bpc'
        self.Pck3_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/pck/pck00010.tpc'

        self.Fk_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/fk/eros_fixed.tf'

        self.Ikgrs_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/grs12.ti'
        self.Ikmsi_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/msi15.ti'
        self.Iknis_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/nis14.ti'
        self.Iknlr_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/nlr04.ti'
        self.Ikxrs_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/ik/xrs12.ti'

        self.SpkPlanet_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/de403s.bsp'
        self.SpkEros_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/eros80.bsp'
        self.SpkEros2_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/erosephem_1999004_2002181.bsp'
        self.SpkMath_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/math9749.bsp'
        self.SpkNearLanded_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/near_eroslanded_nav_v1.bsp'
        self.SpkNearOrbit_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/near_erosorbit_nav_v1.bsp'
        self.SpkStations_url = 'https://naif.jpl.nasa.gov/pub/naif/pds/data/near-a-spice-6-v1.0/nearsp_1000/data/spk/stations.bsp'
        
        # get the path for each kernel
        self.logger.info('Now converting all the URLs to a local path by parsing the URLs')

        self.urlList = [self.Lsk_url, self.Ck_url, self.Sclk_url, self.Pck3_url,
                        self.PckEros1_url, self.PckEros2_url, self.Fk_url,
                        self.Ikgrs_url, self.Ikmsi_url, self.Iknis_url, self.Iknlr_url,
                        self.Ikxrs_url, self.SpkPlanet_url, self.SpkEros_url,
                        self.SpkEros2_url, self.SpkMath_url, self.SpkNearLanded_url,
                        self.SpkNearOrbit_url, self.SpkStations_url]

        self.Lsk = getPathfromUrl(self.Lsk_url)
        self.Ck = getPathfromUrl(self.Ck_url)
        self.Sclk = getPathfromUrl(self.Sclk_url)

        self.PckEros1 = getPathfromUrl(self.PckEros1_url)
        self.PckEros2 = getPathfromUrl(self.PckEros2_url)
        self.Pck3 = getPathfromUrl(self.Pck3_url)

        self.Fk = getPathfromUrl(self.Fk_url)

        self.Ikgrs = getPathfromUrl(self.Ikgrs_url)
        self.Ikmsi = getPathfromUrl(self.Ikmsi_url)
        self.Iknis = getPathfromUrl(self.Iknis_url)
        self.Iknlr = getPathfromUrl(self.Iknlr_url)
        self.Ikxrs = getPathfromUrl(self.Ikxrs_url)

        self.SpkPlanet = getPathfromUrl(self.SpkPlanet_url)
        self.SpkEros = getPathfromUrl(self.SpkEros_url)
        self.SpkEros2 = getPathfromUrl(self.SpkEros2_url)
        self.SpkMath = getPathfromUrl(self.SpkMath_url)
        self.SpkNearLanded = getPathfromUrl(self.SpkNearLanded_url)
        self.SpkNearOrbit = getPathfromUrl(self.SpkNearOrbit_url)
        self.SpkStations = getPathfromUrl(self.SpkStations_url)


        self.kernelList = [self.Lsk, self.Ck, self.Sclk, self.Pck3, self.PckEros1,
                           self.PckEros2, self.Fk, self.Ikgrs, self.Ikmsi, self.Iknis,
                           self.Iknlr, self.Ikxrs, self.SpkPlanet, self.SpkEros,
                           self.SpkEros2, self.SpkMath, self.SpkNearLanded,
                           self.SpkNearOrbit, self.SpkStations]

        self.nameList = [getKernelNameFromUrl(url) for url in self.urlList]
        self.kernelDescription = 'Metal Kernel for 2001 NEAR orbit and landing'

        self.logger.info('Starting the download for NEAR kernels')
        getKernels(self, os.path.realpath(path))

        self.logger.info('Creating a NEAR metakernel')
        self.metakernel = writeMetaKernel(self, 'near2001.tm')

        self.info()

    def info(self):
        """Read and Output info about the loaded kernels
        """
        spice.kclear()

        spice.furnsh(self.metakernel)

        self.spkList = [self.SpkPlanet, self.SpkEros, self.SpkEros2,
                        self.SpkMath, self.SpkNearLanded, self.SpkNearOrbit,
                        self.SpkStations]

        self.ckList = [self.Ck]
        self.pckList = [self.PckEros1, self.PckEros2]

        # check SPK coverage
        self.bodies = {}
        self.bodies_coverage = {}
        for spk in self.spkList:
            idcell = spice.spkobj(spk)
            for code in idcell:
                cover = spice.stypes.SPICEDOUBLE_CELL(1000)
                spice.spkcov(spk, code, cover)
                self.bodies[str(code)] = spice.bodc2n(code)
                self.bodies[spice.bodc2n(code)] = code
                self.bodies_coverage[str(code)] = [x for x in cover]
                self.bodies_coverage[spice.bodc2n(code)] = [x for x in cover]

        # check CK coverage
        self.ckframes = {}
        self.ckframes_coverage = {}
        for ck in self.ckList:
            idcell = spice.ckobj(ck)
            for code in idcell:
                cover = spice.ckcov(ck, code, True, 'SEGMENT', 0.0, 'TDB')
                self.ckframes[str(code)] = spice.frmnam(code)
                self.ckframes[spice.frmnam(code)] = code
                self.ckframes_coverage[str(code)] = [x for x in cover]
                self.ckframes_coverage[spice.frmnam(code)] = [x for x in cover]

        # check pck coverage
        self.pckframes = {}
        for pck in self.pckList:
            ids = spice.stypes.SPICEINT_CELL(1000)
            spice.pckfrm(pck, ids)
            for code in ids:
                self.pckframes[str(code)] = spice.frmnam(code)
                self.pckframes[spice.frmnam(code)] = code

        spice.kclear()


class CassiniKernels(object):
    """List of urls and kernels for the Cassini mission

    More data on Cassini is available:
    https://naif.jpl.nasa.gov/pub/naif/CASSINI/
    """

    def __init__(self, path=cwd):
        """Initialize the Cassini kernel class
        """
        self.Lsk_url = 'https://naif.jpl.nasa.gov/pub/naif/generic_kernels/lsk/a_old_versions/naif0011.tls'
        self.Sclk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/sclk/cas00171.tsc'
        self.Pck_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/pck/cpck09May2017.tpc'
        self.Fk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/fk/release.11/cas_v37.tf'
        self.Ck_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/ck/04135_04171pc_psiv2.bc'
        self.Spk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/981005_PLTEPH-DE405S.bsp'
        self.Ik_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/ik/release.11/cas_iss_v09.ti'
        self.TourSpk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/030201AP_SK_SM546_T45.bsp'
        self.satSpk_url = 'https://naif.jpl.nasa.gov/pub/naif/CASSINI/kernels/spk/020514_SE_SAT105.bsp'

        self.Lsk = getPathfromUrl(self.Lsk_url)
        self.Sclk = getPathfromUrl(self.Sclk_url)
        self.Pck = getPathfromUrl(self.Pck_url)
        self.Fk = getPathfromUrl(self.Fk_url)
        self.Ck = getPathfromUrl(self.Ck_url)
        self.Spk = getPathfromUrl(self.Spk_url)
        self.Ik = getPathfromUrl(self.Ik_url)
        self.TourSpk = getPathfromUrl(self.TourSpk_url)
        self.satSpk = getPathfromUrl(self.satSpk_url)

        self.urlList = [self.Lsk_url, self.Sclk_url, self.Pck_url, self.Fk_url,
                        self.Ck_url, self.Spk_url, self.Ik_url, self.TourSpk_url,
                        self.satSpk_url]
        self.kernelList = [self.Lsk, self.Sclk, self.Pck, self.Fk, self.Ck,
                           self.Spk, self.Ik, self.TourSpk, self.satSpk]
        self.nameList = [getKernelNameFromUrl(url) for url in self.urlList]

        self.kernelDescription = 'Metal Kernel for Cassini Orbiter'
        getKernels(self, os.path.realpath(path))
        self.metakernel = writeMetaKernel(self, 'cassini.tm')


def cleanupKernels(kernelObj=CassiniKernels):
    """Delete all the Kernels
    """
    for kernel in kernelObj.kernelList:
        delete_file(kernel)

def attemptDownload(url, kernelName, targetFileName, num_attempts=5, logger=logging.getLogger(__name__)):
    """Download the file from a specific url
    """
    current_attempt = 0
    while current_attempt < num_attempts:
        try:
            logger.info("Attempting to download kernel: {}".format(kernelName))
            urlretrieve(url, targetFileName)
            break
        except: # TODO Catch URL exception ehre - just a warning also catch a real error if the download fails log it both
            pass

        current_attempt = current_attempt + 1
        logger.info("Attempting to download kernel again...")
        time.sleep(2 + current_attempt)

    if current_attempt >= num_attempts:
        # TODO Add an exception here for a real failure
        logger.error("Error downloading kernel: {}. Check if it exists at url: {}".format(
            kernelName, url))


def getKernels(kernelObj=CassiniKernels, path=cwd):
    """Download all the Kernels
    """

    for url in kernelObj.urlList:
        kernelName = getKernelNameFromUrl(url)
        kernelFile = os.path.join(path, directory, kernelName)

        if not os.path.isfile(kernelFile):
            attemptDownload(url, kernelName, kernelFile, 5)

    return 0


def writeMetaKernel(kernelObj, filename='testKernel.tm'):
    """Write a user defined meta kernel file
    """
    with open(os.path.join(cwd, directory, filename), 'w') as metaKernel:
        metaKernel.write('\\begintext\n\n')
        metaKernel.write('Created: Shankar Kulumani\n')
        if kernelObj.kernelDescription:
            metaKernel.write('Description: {}\n'.format(
                kernelObj.kernelDescription))

        metaKernel.write('\n')

        metaKernel.write('\\begindata\n\n')
        metaKernel.write('PATH_VALUES = (\n')
        metaKernel.write('\'{0}\'\n'.format(os.path.join(cwd, directory)))
        metaKernel.write(')\n\n')

        metaKernel.write('PATH_SYMBOLS = (\n')
        metaKernel.write('\'KERNELS\'\n')
        metaKernel.write(')\n\n')

        metaKernel.write('KERNELS_TO_LOAD = (\n')
        for kernel in kernelObj.nameList:
            metaKernel.write('\'$KERNELS/{0}\'\n'.format(kernel))

        metaKernel.write(')\n')
        metaKernel.close()

    return os.path.join(cwd, directory, filename)


if __name__ == '__main__':
    pass
